/*
 * Copyright (c) 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $FreeBSD: src/sys/pci/if_rl.c,v 1.38.2.7 2001/07/19 18:33:07 wpaul Exp $
 */

/*
 * RealTek 8129/8139 PCI NIC driver
 *
 * Written by Bill Paul <wpaul@ctr.columbia.edu>
 * Electrical Engineering Department
 * Columbia University, New York City
 */

 /*
 * This driver also support Realtek 8139C+, 8110S/SB/SC, RTL8111B/C/CP/D and RTL8101E/8102E/8103E.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sockio.h>
#include <sys/mbuf.h>
#include <sys/malloc.h>
#include <sys/kernel.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <net/if_dl.h>
#include <net/if_media.h>

#include <net/bpf.h>

#include <vm/vm.h>              /* for vtophys */
#include <vm/pmap.h>            /* for vtophys */
#include <machine/clock.h>      /* for DELAY */

#include <machine/bus.h>
#include <machine/resource.h>
#include <sys/bus.h>
#include <sys/rman.h>
#include <sys/endian.h>

#include <pci/if_rlreg.h>
#if OS_VER < VERSION(5,3)
#include <pci/pcireg.h>
#include <pci/pcivar.h>
#include <machine/bus_pio.h>
#include <machine/bus_memio.h>
#else
#include <dev/pci/pcireg.h>
#include <dev/pci/pcivar.h>
#include <sys/module.h>
#endif

#if OS_VER > VERSION(5,9)
#include <sys/cdefs.h>
#include <sys/endian.h>
#include <net/if_types.h>
#include <net/if_vlan_var.h>
#endif


/*
 * Default to using PIO access for this driver. On SMP systems,
 * there appear to be problems with memory mapped mode: it looks like
 * doing too many memory mapped access back to back in rapid succession
 * can hang the bus. I'm inclined to blame this on crummy design/construction
 * on the part of RealTek. Memory mapped mode does appear to work on
 * uniprocessor systems though.
 */
#define RL_USEIOSPACE


#ifndef lint
static const char rcsid[] =
  "$FreeBSD: src/sys/pci/if_rl.c,v 1.38.2.7 2001/07/19 18:33:07 wpaul Exp $";
#endif

#define EE_SET(x)					\
	CSR_WRITE_1(sc, RL_EECMD,			\
		CSR_READ_1(sc, RL_EECMD) | x)

#define EE_CLR(x)					\
	CSR_WRITE_1(sc, RL_EECMD,			\
		CSR_READ_1(sc, RL_EECMD) & ~x)

#ifdef RL_USEIOSPACE
#define RL_RES			SYS_RES_IOPORT
#define RL_RID			RL_PCI_LOIO
#else
#define RL_RES			SYS_RES_MEMORY
#define RL_RID			RL_PCI_LOMEM
#endif

/*
 * Various supported device vendors/types and their names.
 */
static struct rl_type rl_devs[] = {
	{ RT_VENDORID, RT_DEVICEID_8129,
		"Realtek PCI FE Family Controller" },
	{ RT_VENDORID, RT_DEVICEID_8139,
		"Realtek PCI FE Family Controller" },
	{ RT_VENDORID, RT_DEVICEID_8169,
		"Realtek PCI GBE Family Controller" },
	{ RT_VENDORID, RT_DEVICEID_8169SC,
		"Realtek PCI GBE Family Controller" },
	{ RT_VENDORID, RT_DEVICEID_8168B,
		"Realtek PCIe GBE Family Controller" },
	{ RT_VENDORID, RT_DEVICEID_8101E,
		"Realtek PCIe FE Family Controller" },
	{ ACCTON_VENDORID, ACCTON_DEVICEID_5030,
		"Accton MPX 5030/5038 10/100BaseTX" },
	{ DELTA_VENDORID, DELTA_DEVICEID_8139,
		"Delta Electronics 8139 10/100BaseTX" },
	{ ADDTRON_VENDORID, ADDTRON_DEVICEID_8139,
		"Addtron Technolgy 8139 10/100BaseTX" },
	{ DLINK_VENDORID, DLINK_DEVICEID_530TXPLUS,
		"D-Link DFE-530TX+ 10/100BaseTX" },
	{ 0, 0, NULL }
};

static int	rl_probe			__P((device_t));
static int	rl_attach			__P((device_t));
static int	rl_detach			__P((device_t));
static void rl_shutdown				__P((device_t));

static void MP_WritePhyUshort			__P((struct rl_softc*, u_int8_t, u_int16_t));
static u_int16_t MP_ReadPhyUshort		__P((struct rl_softc *,u_int8_t));
static void MP_WriteEPhyUshort			__P((struct rl_softc*, u_int8_t, u_int16_t));
static u_int16_t MP_ReadEPhyUshort		__P((struct rl_softc *,u_int8_t));
static u_int8_t MP_ReadEfuse			__P((struct rl_softc *,u_int16_t));

static void rl_8169s_init			__P((struct rl_softc *));
static void rl_init				__P((void *));
static int 	rl_var_init			__P((struct rl_softc *));
static void rl_reset				__P((struct rl_softc *));
static void rl_stop				__P((struct rl_softc *));

static void rl_start				__P((struct ifnet *));
static int rl_encap				__P((struct rl_softc *, struct mbuf * ));
static void WritePacket				__P((struct rl_softc *, caddr_t, int, int, int));
static int CountFreeTxDescNum			__P((struct rl_descriptor));
static int CountMbufNum				__P((struct mbuf *, int *));
static void rl_txeof				__P((struct rl_softc *));

static void	rl_rxeof			__P((struct rl_softc *));

static void rl_intr				__P((void *));
static u_int8_t rl_calchash			__P((caddr_t));
static void rl_setmulti				__P((struct rl_softc *));
static int	rl_ioctl			__P((struct ifnet *, u_long, caddr_t));
static void rl_tick				__P((void *));
#if OS_VER<VERSION(7,0)
static void rl_watchdog				__P((struct ifnet *));
#endif

static int	rl_ifmedia_upd			__P((struct ifnet *));
static void rl_ifmedia_sts			__P((struct ifnet *, struct ifmediareq *));

static void rl_eeprom_ShiftOutBits		__P((struct rl_softc *, int, int));
static u_int16_t rl_eeprom_ShiftInBits		__P((struct rl_softc *));
static void rl_eeprom_EEpromCleanup		__P((struct rl_softc *));
static void rl_eeprom_getword			__P((struct rl_softc *, int, u_int16_t *));
static void rl_read_eeprom			__P((struct rl_softc *, caddr_t, int, int, int));

static void rl_phy_power_up(device_t dev);
#if 0
static void rl_phy_power_down(device_t dev);
#endif

static void rl_set_speed_and_duplex(struct rl_softc *sc, int gbcr, int anar, int bmcr);

static device_method_t rl_methods[] = {
	/* Device interface */
	DEVMETHOD(device_probe,		rl_probe),
	DEVMETHOD(device_attach,	rl_attach),
	DEVMETHOD(device_detach,	rl_detach),
	DEVMETHOD(device_shutdown,	rl_shutdown),
	{ 0, 0 }
};

static driver_t rl_driver = {
	"rl",
	rl_methods,
	sizeof(struct rl_softc)
};

static devclass_t rl_devclass;

DRIVER_MODULE(if_rl, pci, rl_driver, rl_devclass, 0, 0);

static void rl_phy_power_up(dev)
	device_t		dev;
{
	struct rl_softc		*sc;

	sc = device_get_softc(dev);

	switch (sc->rl_type) {
	case MACFG_4:
	case MACFG_5:
	case MACFG_6:
	case MACFG_21:
	case MACFG_22:
	case MACFG_23:
	case MACFG_24:
	case MACFG_25:
	case MACFG_26:
	case MACFG_27:
	case MACFG_28:
	case MACFG_31:
	case MACFG_32:
	case MACFG_33:
		MP_WritePhyUshort(sc, 0x1f, 0x0000);
		MP_WritePhyUshort(sc, 0x0e, 0x0000);
		break;
	case MACFG_11:
	case MACFG_12:
	case MACFG_13:
	case MACFG_14:
	case MACFG_15:
	case MACFG_16:
		break;
	};
}

#if 0
static void rl_phy_power_down(dev)
	device_t		dev;
{
	struct rl_softc		*sc;

	sc = device_get_softc(dev);

	switch (sc->rl_type) {
	case MACFG_21:
	case MACFG_22:
	case MACFG_23:
	case MACFG_24:
	case MACFG_25:
	case MACFG_26:
	case MACFG_27:
	case MACFG_28:
	case MACFG_31:
	case MACFG_32:
	case MACFG_33:
		MP_WritePhyUshort(sc, 0x1f, 0x0000);
		MP_WritePhyUshort(sc, 0x0e, 0x0200);
		MP_WritePhyUshort(sc, 0x00, 0x0000);
		break;
	case MACFG_4:
	case MACFG_5:
	case MACFG_6:
	case MACFG_11:
	case MACFG_12:
	case MACFG_13:
	case MACFG_14:
	case MACFG_16:
		MP_WritePhyUshort(sc, 0x00, 0x0000);
		break;
	}
}
#endif

/*
 * Probe for a RealTek 8129/8139 chip. Check the PCI vendor and device
 * IDs against our list and return a device name if we find a match.
 */
static int rl_probe(dev)	/* Search for Realtek NIC chip */
	device_t		dev;
{
	struct rl_type		*t; t = rl_devs;
	while(t->rl_name != NULL) {
		if ((pci_get_vendor(dev) == t->rl_vid) &&
		    (pci_get_device(dev) == t->rl_did))
		{
			device_set_desc(dev, t->rl_name);
			return(0);
		}
		t++;
	}

	return(ENXIO);
}

/*
 * Attach the interface. Allocate softc structures, do ifmedia
 * setup and ethernet/BPF attach.
 */
static int rl_attach(device_t dev)
{
	/*int			s;*/
	u_char			eaddr[ETHER_ADDR_LEN];
	u_int32_t		command;
	struct rl_softc		*sc;
	struct ifnet		*ifp;
	u_int16_t		rl_did = 0;
	int			unit, error = 0, rid, i;
	int			mac_version;
	int			mode;
	u_int8_t		data8;

	/*s = splimp();*/

	sc = device_get_softc(dev);
	unit = device_get_unit(dev);
	bzero(sc, sizeof(struct rl_softc));
	RL_LOCK_INIT(sc,device_get_nameunit(dev));

	sc->driver_detach = 0;

	pci_enable_busmaster(dev);
	/*
	 * Handle power management nonsense.
	 */
	command = pci_read_config(dev, RL_PCI_CAPID, 4) & 0x000000FF;
	if (command == 0x01) {

		command = pci_read_config(dev, RL_PCI_PWRMGMTCTRL, 4);
		if (command & RL_PSTATE_MASK) {
			u_int32_t		iobase, membase, irq;

			/* Save important PCI config data. */
			iobase = pci_read_config(dev, RL_PCI_LOIO, 4);
			membase = pci_read_config(dev, RL_PCI_LOMEM, 4);
			irq = pci_read_config(dev, RL_PCI_INTLINE, 4);

			/* Reset the power state. */
			printf("rl%d: chip is is in D%d power mode "
			"-- setting to D0\n", unit, command & RL_PSTATE_MASK);
			command &= 0xFFFFFFFC;
			pci_write_config(dev, RL_PCI_PWRMGMTCTRL, command, 4);

			/* Restore PCI config data. */
			pci_write_config(dev, RL_PCI_LOIO, iobase, 4);
			pci_write_config(dev, RL_PCI_LOMEM, membase, 4);
			pci_write_config(dev, RL_PCI_INTLINE, irq, 4);
		}
	}

	/*
	 * Map control/status registers.
	 */
	command = pci_read_config(dev, PCIR_COMMAND, 4);
	command |= (PCIM_CMD_PORTEN|PCIM_CMD_MEMEN|PCIM_CMD_BUSMASTEREN);
	pci_write_config(dev, PCIR_COMMAND, command, 4);
	command = pci_read_config(dev, PCIR_COMMAND, 4);

#ifdef RL_USEIOSPACE
	if (!(command & PCIM_CMD_PORTEN)) {
		printf("rl%d: failed to enable I/O ports!\n", unit);
		error = ENXIO;
		goto fail;
	}
#else
	if (!(command & PCIM_CMD_MEMEN)) {
		printf("rl%d: failed to enable memory mapping!\n", unit);
		error = ENXIO;
		goto fail;
	}
#endif

	rid = RL_RID;
	sc->rl_res = bus_alloc_resource(dev, RL_RES, &rid,
	    0, ~0, 1, RF_ACTIVE);

	if (sc->rl_res == NULL) {
		printf ("rl%d: couldn't map ports/memory\n", unit);
		error = ENXIO;
		goto fail;
	}

	sc->rl_btag = rman_get_bustag(sc->rl_res);
	sc->rl_bhandle = rman_get_bushandle(sc->rl_res);

	rid = 0;
	sc->rl_irq = bus_alloc_resource(dev, SYS_RES_IRQ, &rid, 0, ~0, 1,
	    RF_SHAREABLE | RF_ACTIVE);

	if (sc->rl_irq == NULL) {
		printf("rl%d: couldn't map interrupt\n", unit);
		error = ENXIO;
		goto fail;
	}

	callout_handle_init(&sc->rl_stat_ch);

	/*
	 * Reset the adapter. Only take the lock here as it's needed in
	 * order to call rl_reset().
	 */
	RL_LOCK(sc);
	rl_reset(sc);
	RL_UNLOCK(sc);

	/* Get station address from the EEPROM. */
	for (i = 0; i < ETHER_ADDR_LEN; i++)
		eaddr[i] = CSR_READ_1(sc, RL_IDR0 + i);

	/*
	 * A RealTek chip was detected. Inform the world.
	 */
	device_printf(dev,"version:1.79\n");
	device_printf(dev,"Ethernet address: %6D\n", eaddr, ":");
	printf("\nThis product is covered by one or more of the following patents: US5,307,459, US5,434,872, US5,732,094, US6,570,884, US6,115,776, and US6,327,625.\n");

	sc->rl_unit = unit;

#if OS_VER<VERSION(6,0)
	bcopy(eaddr, (char *)&sc->arpcom.ac_enaddr, ETHER_ADDR_LEN);
#endif

	/*
	 * Now read the exact device type from the EEPROM to find
	 * out if it's an 8129 or 8139.
	 */
	rl_read_eeprom(sc, (caddr_t)&rl_did, RL_EE_PCI_DID, 1, 0);
	mac_version = (CSR_READ_4(sc, RL_TXCFG) & 0xFC800000);
	mode = (CSR_READ_4(sc, RL_TXCFG) & 0x00700000);

	if (rl_did == RT_DEVICEID_8139 ||
	    rl_did == ACCTON_DEVICEID_5030 ||
	    rl_did == DELTA_DEVICEID_8139 ||
	    rl_did == ADDTRON_DEVICEID_8139 ||
	    rl_did == DLINK_DEVICEID_530TXPLUS) {
		sc->rl_type = RL_8139;
	} else if (rl_did == RT_DEVICEID_8129) {
		sc->rl_type = RL_8129;
	}
	else
	{
		switch(mac_version)
		{
			case 0x00800000:
			case 0x04000000:
				sc->rl_type = MACFG_3;
				sc->mac_style = MAC_STYLE_1;
				break;

			case 0x10000000:
				sc->rl_type = MACFG_4;
				sc->mac_style = MAC_STYLE_1;
				break;

			case 0x18000000:
				sc->rl_type = MACFG_5;
				sc->mac_style = MAC_STYLE_1;
				break;

			case 0x98000000:
				sc->rl_type = MACFG_6;
				sc->mac_style = MAC_STYLE_1;
				break;

			case 0x34000000:
			case 0xB4000000:
				sc->mac_style = MAC_STYLE_1;

				if (mode == 0x00000000) {
					sc->rl_type = MACFG_11;
				} else if (mode == 0x00200000) {
					sc->rl_type = MACFG_12;
				} else if (mode == 0x00300000) {
					sc->rl_type = MACFG_13;
				} else {
					sc->rl_type = MACFG_13;
				}
				break;

			case 0x34800000:
			case 0x24800000:
				sc->mac_style = MAC_STYLE_2;

				if (mode == 0x00100000) {
					sc->rl_type = MACFG_14;
				} else if (mode == 0x00200000) {
					sc->rl_type = MACFG_15;
				} else if (mode == 0x00400000) {
					sc->rl_type = MACFG_16;
				} else if (mode == 0x00500000) {
					sc->rl_type = MACFG_17;
				} else {
					sc->rl_type = MACFG_17;
				}
				break;

			case 0x24000000:
				sc->rl_type = MACFG_18;
				sc->mac_style = MAC_STYLE_2;
				break;

			case 0x30000000:
				sc->rl_type = MACFG_21;
				sc->mac_style = MAC_STYLE_1;
				break;

			case 0x38000000:
			case 0xB8000000:
				sc->mac_style = MAC_STYLE_1;

				if (mode == 0x00000000) {
					sc->rl_type = MACFG_22;
				} else if (mode == 0x00500000) {
					sc->rl_type = MACFG_23;
				} else {
					sc->rl_type = MACFG_23;
				}
				break;

			case 0x3C000000:
				sc->mac_style = MAC_STYLE_2;

				if (mode == 0x00000000) {
					sc->rl_type = MACFG_24;
				} else if (mode == 0x00200000) {
					sc->rl_type = MACFG_25;
				} else if (mode == 0x00400000) {
					sc->rl_type = MACFG_26;
				} else {
					sc->rl_type = MACFG_26;
				}
				break;

			case 0x3C800000:
				sc->mac_style = MAC_STYLE_2;

				if (mode == 0x00100000) {
					sc->rl_type = MACFG_27;
				} else if (mode == 0x00300000) {
					sc->rl_type = MACFG_28;
				} else {
					sc->rl_type = MACFG_28;
				}
				break;

			case 0x28000000:
				sc->mac_style = MAC_STYLE_2;

				if (mode == 0x00100000) {
					sc->rl_type = MACFG_31;
				} else if (mode == 0x00200000) {
					sc->rl_type = MACFG_32;
				} else if (mode == 0x00300000) {
					sc->rl_type = MACFG_33;
				} else {
					sc->rl_type = MACFG_33;
				}
				break;

			default:
				device_printf(dev,"unknown device ID: %x\n", rl_did);
				device_printf(dev,"mac_vesion: %x\n", mac_version);
				error = ENXIO;
				goto fail;
		}
	}

	if(sc->rl_type == RL_8139)	/* distinct between 8139c & 8139c+ */
	{
		if(pci_read_config(dev,RL_PCI_REVISION_ID,1)==RL_CPlusMode)
		{
			sc->rl_bIsCPlusMode=TRUE;
			DBGPRINT(unit, "Rtl8139C+ detected !");
		}
		else
		{
			sc->rl_bIsCPlusMode=FALSE;
			DBGPRINT(unit, "Rtl8139 detected !");
		}
	}
	else if(sc->rl_type==MACFG_3)
	{	/* Change PCI Latency time*/
		pci_write_config(dev, RL_PCI_LATENCY_TIMER, 0x40, 1);
	}

	if ((sc->rl_type == RL_8139 && sc->rl_bIsCPlusMode) ||
	    (sc->mac_style == MAC_STYLE_1) ||
	    (sc->mac_style == MAC_STYLE_2)) {
		sc->DESCRIPTOR_MODE = TRUE;
		DBGPRINT(unit, "Descriptor mode turn ON !\n");
	} else {
		DBGPRINT(unit, "Descriptor mode turn OFF !\n");
	}

	if (sc->mac_style == MAC_STYLE_2) {
		data8 = pci_read_config(dev, 0x79, 1);
		data8 &= ~0x70;
		data8 |= 0x50;
		pci_write_config(dev, RL_PCI_PWRMGMTCTRL, data8, 1);

		pci_write_config(dev, 0x80, 0x00, 1);
	}

	if(!sc->DESCRIPTOR_MODE)
	{
		sc->rl_cdata.rl_rx_buf = contigmalloc(RL_RXBUFLEN + 1518, M_DEVBUF,
			M_NOWAIT, 0, 0xffffffff, PAGE_SIZE, 0);

		if (sc->rl_cdata.rl_rx_buf == NULL) {
			printf("rl%d: no memory for list buffers!\n", unit);
			error = ENXIO;
			goto fail;
		}

		/* Leave a few bytes before the start of the RX ring buffer. */
		sc->rl_cdata.rl_rx_buf_ptr = sc->rl_cdata.rl_rx_buf;
		sc->rl_cdata.rl_rx_buf += sizeof(u_int64_t);
	}
	else	/* Descriptor Mode */
	{
		sc->rl_desc.rx_desc=contigmalloc(sizeof(union RxDesc)*RL_RX_BUF_NUM, M_DEVBUF,
			M_NOWAIT|M_ZERO, 0, BUS_SPACE_MAXADDR, RL_DESC_ALIGN, 0);
		sc->rl_desc.tx_desc=contigmalloc(sizeof(union TxDesc)*RL_TX_BUF_NUM, M_DEVBUF,
			M_NOWAIT|M_ZERO, 0, BUS_SPACE_MAXADDR, RL_DESC_ALIGN, 0);
		for(i=0;i<RL_RX_BUF_NUM;i++)
		{
			sc->rl_desc.rx_buf[i] = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR);
			if(!sc->rl_desc.rx_buf[i])
			{
				device_printf(dev, "m_getcl fail!!!\n");
				error = ENXIO;
				goto fail;

			}
		}
	}

	sc->rl_8169_MacVersion=(CSR_READ_4(sc, RL_TXCFG)&0x7c800000)>>25;		/* Get bit 26~30 	*/
	sc->rl_8169_MacVersion|=((CSR_READ_4(sc, RL_TXCFG)&0x00800000)!=0 ? 1:0);	/* Get bit 23 		*/
	DBGPRINT1(sc->rl_unit,"8169 Mac Version %d",sc->rl_8169_MacVersion);

	/* Rtl8169s single chip detected */
	if(sc->rl_8169_MacVersion > 0) {
		sc->rl_8169_PhyVersion=(MP_ReadPhyUshort(sc, 0x03)&0x000f);
		DBGPRINT1(sc->rl_unit,"8169 Phy Version %d",sc->rl_8169_PhyVersion);

		if(sc->rl_8169_MacVersion == 1) {
			CSR_WRITE_1(sc, 0x82, 0x01);
			MP_WritePhyUshort(sc, 0x0b, 0x00);
		}

		rl_phy_power_up(dev);
		rl_8169s_init(sc);
	}

#if OS_VER<VERSION(6,0)
	ifp = &sc->arpcom.ac_if;
#else
	ifp = sc->rl_ifp = if_alloc(IFT_ETHER);
	if (ifp == NULL) {
		device_printf(dev, "can not if_alloc()\n");
		error = ENOSPC;
		goto fail;
	}
#endif
	ifp->if_softc = sc;
#if OS_VER < VERSION(5,3)
	ifp->if_unit = unit;
	ifp->if_name = "rl";
#else
	if_initname(ifp, device_get_name(dev), device_get_unit(dev));
#endif
	ifp->if_mtu = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST | IFF_SIMPLEX | IFF_MULTICAST;
	ifp->if_ioctl = rl_ioctl;
	ifp->if_output = ether_output;
	ifp->if_start = rl_start;
#if OS_VER<VERSION(7,0)
	ifp->if_watchdog = rl_watchdog;
#endif
	ifp->if_init = rl_init;
	switch(sc->rl_type)
	{
		case RL_8129:
		case RL_8139:
		case MACFG_11:
		case MACFG_12:
		case MACFG_13:
		case MACFG_14:
		case MACFG_15:
		case MACFG_16:
		case MACFG_17:
		case MACFG_18:
			ifp->if_baudrate = 100000000;
			break;

		default:
			ifp->if_baudrate = 1000000000;
			break;
	}
	IFQ_SET_MAXLEN(&ifp->if_snd, IFQ_MAXLEN);
	ifp->if_snd.ifq_drv_maxlen = IFQ_MAXLEN;
	IFQ_SET_READY(&ifp->if_snd);

	/*
	 * Call MI attach routine.
	 */
/*#if OS_VER < VERSION(5,1)*/
#if OS_VER < VERSION(4,9)
	ether_ifattach(ifp, ETHER_BPF_SUPPORTED);
#else
	ether_ifattach(ifp, eaddr);
#endif


#if OS_VER<VERSION(7,0)
	error = bus_setup_intr(dev, sc->rl_irq, INTR_TYPE_NET,
	    rl_intr, sc, &sc->rl_intrhand);
#else
	error = bus_setup_intr(dev, sc->rl_irq, INTR_TYPE_NET|INTR_MPSAFE,
			NULL, rl_intr, sc, &sc->rl_intrhand);
#endif

	if (error) {
		device_printf(dev,"couldn't set up irq\n");
		goto fail;
	}

	/*
	 * Specify the media types supported by this adapter and register
	 * callbacks to update media and link information
	 */
	ifmedia_init(&sc->media, IFM_IMASK, rl_ifmedia_upd,
		     rl_ifmedia_sts);
	/*if (sc->hw.media_type == em_media_type_fiber)
	{
		ifmedia_add(&sc->media, IFM_ETHER | IFM_1000_SX | IFM_FDX,
			    0, NULL);
		ifmedia_add(&sc->media, IFM_ETHER | IFM_1000_SX,
			    0, NULL);
	} else */
	{
		ifmedia_add(&sc->media, IFM_ETHER | IFM_10_T, 0, NULL);
		ifmedia_add(&sc->media, IFM_ETHER | IFM_10_T | IFM_FDX,
			    0, NULL);
		ifmedia_add(&sc->media, IFM_ETHER | IFM_100_TX,
			    0, NULL);
		ifmedia_add(&sc->media, IFM_ETHER | IFM_100_TX | IFM_FDX,
			    0, NULL);
/*#if OS_VER < VERSION(5,1)*/
#if OS_VER < VERSION(4,9)
		ifmedia_add(&sc->media, IFM_ETHER | IFM_1000_TX | IFM_FDX,
			    0, NULL);
		ifmedia_add(&sc->media, IFM_ETHER | IFM_1000_TX, 0, NULL);
#else
		ifmedia_add(&sc->media, IFM_ETHER | IFM_1000_T | IFM_FDX,
			    0, NULL);
		ifmedia_add(&sc->media, IFM_ETHER | IFM_1000_T, 0, NULL);
#endif
	}
	ifmedia_add(&sc->media, IFM_ETHER | IFM_AUTO, 0, NULL);
	ifmedia_set(&sc->media, IFM_ETHER | IFM_AUTO);

fail:
	if (error)
		rl_detach(dev);

	return(error);
}

static int rl_detach(device_t dev)
{
	struct rl_softc		*sc;
	struct ifnet		*ifp;
	/*int			s;*/
	int			i;

	/*s = splimp();*/

	sc = device_get_softc(dev);

	if (sc->rl_intrhand)
		bus_teardown_intr(dev, sc->rl_irq, sc->rl_intrhand);
	if (sc->rl_irq)
		bus_release_resource(dev, SYS_RES_IRQ, 0, sc->rl_irq);
	if (sc->rl_res)
		bus_release_resource(dev, RL_RES, RL_RID, sc->rl_res);

#if OS_VER<VERSION(6,0)
	ifp = &sc->arpcom.ac_if;
#else
	ifp = sc->rl_ifp;
#endif

	/* These should only be active if attach succeeded */
	if (device_is_attached(dev))
	{
		RL_LOCK(sc);
		rl_stop(sc);
		RL_UNLOCK(sc);
#if OS_VER < VERSION(4,9)
		ether_ifdetach(ifp, ETHER_BPF_SUPPORTED);
#else
		ether_ifdetach(ifp);
#endif
	}
	sc->driver_detach = 1;

#if OS_VER>=VERSION(6,0)
	if (ifp)
		if_free(ifp);
#endif
	bus_generic_detach(dev);

	if(!sc->DESCRIPTOR_MODE)
	{
		contigfree(sc->rl_cdata.rl_rx_buf, RL_RXBUFLEN + 32, M_DEVBUF);
	}
	else
	{
		for(i=0;i<RL_RX_BUF_NUM;i++)
		{
			if(sc->rl_desc.rx_buf[i]!=NULL)
			{
				m_freem(sc->rl_desc.rx_buf[i]);
			}
		}
		contigfree(sc->rl_desc.rx_desc, sizeof(union RxDesc)*RL_RX_BUF_NUM, M_DEVBUF);
		contigfree(sc->rl_desc.tx_desc, sizeof(union TxDesc)*RL_TX_BUF_NUM, M_DEVBUF);
	}

	/*splx(s);*/
	RL_LOCK_DESTROY(sc);

	return(0);
}

/*
 * Stop all chip I/O so that the kernel's probe routines don't
 * get confused by errant DMAs when rebooting.
 */
static void rl_shutdown(dev)	/* The same with rl_stop(sc) */
	device_t		dev;
{
	struct rl_softc		*sc;

	sc = device_get_softc(dev);

	RL_LOCK(sc);
	rl_stop(sc);
	RL_UNLOCK(sc);

	return;
}

static void rl_init(void *xsc)		/* Software & Hardware Initialize */
{
	struct rl_softc		*sc = xsc;
#if OS_VER<VERSION(6,0)
	struct ifnet		*ifp = &sc->arpcom.ac_if;
	int			i;
	/*u_int8_t		ldps = 0;*/
#else
	struct ifnet		*ifp = sc->rl_ifp;
/*	int			s;*/
#endif
	u_int32_t		rxcfg = 0;

	u_int16_t		data16 = 0;

	/*s = splimp();*/
	RL_LOCK(sc);

/*	RL_LOCK_ASSERT(sc);*/

	/*mii = device_get_softc(sc->rl_miibus);*/

	/*
	 * Cancel pending I/O and free all RX/TX buffers.
	 */
	rl_stop(sc);

	/* Init our MAC address */
#if OS_VER<VERSION(6,0)
	for (i = 0; i < ETHER_ADDR_LEN; i++) {
		CSR_WRITE_1(sc, RL_IDR0 + i, sc->arpcom.ac_enaddr[i]);
	}
#elif OS_VER<VERSION(7,0)
	CSR_WRITE_1(sc, RL_EECMD, RL_EEMODE_WRITECFG);
	CSR_WRITE_STREAM_4(sc, RL_IDR0,
	    *(u_int32_t *)(&IFP2ENADDR(sc->rl_ifp)[0]));
	CSR_WRITE_STREAM_4(sc, RL_IDR4,
	    *(u_int32_t *)(&IFP2ENADDR(sc->rl_ifp)[4]));
	CSR_WRITE_1(sc, RL_EECMD, RL_EEMODE_OFF);
#else
	CSR_WRITE_1(sc, RL_EECMD, RL_EEMODE_WRITECFG);
	CSR_WRITE_STREAM_4(sc, RL_IDR0,
	    *(u_int32_t *)(&IF_LLADDR(sc->rl_ifp)[0]));
	CSR_WRITE_STREAM_4(sc, RL_IDR4,
	    *(u_int32_t *)(&IF_LLADDR(sc->rl_ifp)[4]));
	CSR_WRITE_1(sc, RL_EECMD, RL_EEMODE_OFF);
#endif

	/* Init descriptors. */
	rl_var_init(sc);

	/*disable Link Down Power Saving(non-LDPS)*/
	/*CSR_WRITE_1(sc, RL_LDPS, 0x05);*/
	/*ldps=CSR_READ_1(sc, RL_LDPS);*/

	/* Set the individual bit to receive frames for this host only. */
	rxcfg = CSR_READ_4(sc, RL_RXCFG);
	rxcfg |= RL_RXCFG_RX_INDIV;

	/* If we want promiscuous mode, set the allframes bit. */
	if (ifp->if_flags & IFF_PROMISC) {
		rxcfg |= RL_RXCFG_RX_ALLPHYS;
		CSR_WRITE_4(sc, RL_RXCFG, rxcfg);
	} else {
		rxcfg &= ~RL_RXCFG_RX_ALLPHYS;
		CSR_WRITE_4(sc, RL_RXCFG, rxcfg);
	}

	/*
	 * Set capture broadcast bit to capture broadcast frames.
	 */
	if (ifp->if_flags & IFF_BROADCAST) {
		rxcfg |= RL_RXCFG_RX_BROAD;
		CSR_WRITE_4(sc, RL_RXCFG, rxcfg);
	} else {
		rxcfg &= ~RL_RXCFG_RX_BROAD;
		CSR_WRITE_4(sc, RL_RXCFG, rxcfg);
	}

	/*
	 * Program the multicast filter, if necessary.
	 */
	rl_setmulti(sc);

	CSR_WRITE_2(sc,RL_CPCR,0x2082);
	CSR_WRITE_2(sc,RL_IM,0x5151);

	/*
	 * Enable interrupts.
	 */
	CSR_WRITE_2(sc, RL_IMR, RL_INTRS);

	sc->rl_txrev = RL_TX_REV;

	/* Start RX/TX process. */
	CSR_WRITE_4(sc, RL_MISSEDPKT, 0);

	/* Enable receiver and transmitter. */
/*	CSR_WRITE_1(sc, RL_COMMAND, RL_CMD_TX_ENB|RL_CMD_RX_ENB);*/

	if(sc->DESCRIPTOR_MODE)
	{
		CSR_WRITE_1(sc, RL_EECMD, RL_EEMODE_WRITECFG);
		switch(sc->rl_type)
		{
			case MACFG_3:
			case MACFG_4:
				CSR_WRITE_2(sc, RL_CPlusCmd, 0x0003|((sc->rl_type==MACFG_3 && sc->rl_8169_MacVersion==1) ? 0x4008:0));
				break;

			case MACFG_5:
				if(CSR_READ_1(sc, RL_CFG2)&1)
				{
					CSR_WRITE_4(sc, 0x7C, 0x000FFFFF);
				}
				else
				{
					CSR_WRITE_4(sc, 0x7C, 0x000FFF00);
				}
				CSR_WRITE_2(sc, RL_CPlusCmd,0x0008);
				CSR_WRITE_2(sc, 0xe2,0x0000);
				break;

			case MACFG_6:
				if(CSR_READ_1(sc, RL_CFG2)&1)
				{
					CSR_WRITE_4(sc, 0x7C, 0x003FFFFF);
				}
				else
				{
					CSR_WRITE_4(sc, 0x7C, 0x003FFF00);
				}
				CSR_WRITE_2(sc, RL_CPlusCmd,0x0008);
				CSR_WRITE_2(sc, 0xe2,0x0000);
				break;

			case MACFG_14:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				/*set mac register offset 0xd1 to 0x10*/
				CSR_WRITE_1(sc, RL_DBG_reg, 0x10);

				MP_WriteEPhyUshort(sc, 0x03, 0xC2F9);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05EF);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);
				break;

			case MACFG_15:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				/*set mac register offset 0xd1 to 0x10*/
				CSR_WRITE_1(sc, RL_DBG_reg, 0x10);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05EF);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);
				break;

			case MACFG_16:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				/*set mac register offset 0xd1 to 0x10*/
				CSR_WRITE_1(sc, RL_DBG_reg, 0x10);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05EF);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);

				MP_WriteEPhyUshort(sc, 0x03, 0xC4D9);
				MP_WriteEPhyUshort(sc, 0x06, 0xAF35);
				break;

			case MACFG_17:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				/*set mac register offset 0xd1 to 0x10*/
				CSR_WRITE_1(sc, RL_DBG_reg, 0x10);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05EF);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);

				MP_WriteEPhyUshort(sc, 0x19, 0xECF2);
				break;

			case MACFG_18:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				/*set mac register offset 0xd1 to 0x18*/
				CSR_WRITE_1(sc, RL_DBG_reg, 0x18);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05EF);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);

				MP_WriteEPhyUshort(sc, 0x06, 0xAF25);
				MP_WriteEPhyUshort(sc, 0x07, 0x8E68);
				break;

			case MACFG_24:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				/*set mac register offset 0xd1 to 0xf8*/
				CSR_WRITE_1(sc, RL_DBG_reg, 0xF8);

				data16 = MP_ReadEPhyUshort(sc, 0x02) & ~0x1800;
				data16 |= 0x1000;
				MP_WriteEPhyUshort(sc, 0x02, data16);

				data16 = MP_ReadEPhyUshort(sc, 0x03) | 0x0002;
				MP_WriteEPhyUshort(sc, 0x03, data16);

				data16 = MP_ReadEPhyUshort(sc, 0x06) & ~0x0080;
				MP_WriteEPhyUshort(sc, 0x06, data16);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);
				break;

			case MACFG_25:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);

				data16 = MP_ReadEPhyUshort(sc, 0x01) | 0x0001;
				MP_WriteEPhyUshort(sc, 0x01, data16);

				data16 = MP_ReadEPhyUshort(sc, 0x03) & ~0x0620;
				data16 |= 0x0220;
				MP_WriteEPhyUshort(sc, 0x03, data16);
				break;

			case MACFG_26:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
//				CSR_WRITE_2(sc, RL_CPlusCmd, CSR_READ_2(sc, RL_CPlusCmd) & ~0xDF9C);
				break;

			case MACFG_27:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);
				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);

				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				break;

			case MACFG_28:
				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3F000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);
				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);

				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				break;

			case MACFG_31:
				MP_WriteEPhyUshort(sc, 0x01, 0x7C7D);
				MP_WriteEPhyUshort(sc, 0x02, 0x091F);
				MP_WriteEPhyUshort(sc, 0x06, 0xB271);
				MP_WriteEPhyUshort(sc, 0x07, 0xCE00);

				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3f000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~0x11);
				CSR_WRITE_1(sc, RL_DBG_reg, CSR_READ_1(sc, RL_DBG_reg)|0x82);

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
				break;

			case MACFG_32:
				MP_WriteEPhyUshort(sc, 0x01, 0x7C7D);
				MP_WriteEPhyUshort(sc, 0x02, 0x091F);
				MP_WriteEPhyUshort(sc, 0x03, 0xC5BA);
				MP_WriteEPhyUshort(sc, 0x06, 0xB279);
				MP_WriteEPhyUshort(sc, 0x07, 0xAF00);
				MP_WriteEPhyUshort(sc, 0x1E, 0xB8EB);

				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3f000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
				break;

			case MACFG_33:
				MP_WriteEPhyUshort(sc, 0x01, 0x6C7F);
				MP_WriteEPhyUshort(sc, 0x02, 0x011F);
				MP_WriteEPhyUshort(sc, 0x03, 0xC1B2);
				MP_WriteEPhyUshort(sc, 0x1A, 0x0546);
				MP_WriteEPhyUshort(sc, 0x1C, 0x80C4);
				MP_WriteEPhyUshort(sc, 0x1D, 0x78E4);
				MP_WriteEPhyUshort(sc, 0x0A, 0x8100);

				/*set configuration space offset 0x70f to 0x3f*/
				CSR_WRITE_4(sc, RL_CSIDR, 0x3f000000);
				CSR_WRITE_4(sc, RL_CSIAR, 0x8000870C);

				CSR_WRITE_1(sc, RL_CFG1, (CSR_READ_1(sc, RL_CFG1)&0xC0)|0x1F);
				CSR_WRITE_2(sc, RL_RxMaxSize, 0x05F3);
				CSR_WRITE_1(sc, RL_CFG3, CSR_READ_1(sc, RL_CFG3) & ~(1 << 0));
				CSR_WRITE_1(sc, RL_DBG_reg, CSR_READ_1(sc, RL_DBG_reg)|0x82);

				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2020);
				break;

			default:
				CSR_WRITE_2 (sc, RL_CPlusCmd,0x2000);
				break;
		}
		CSR_WRITE_1(sc, RL_EECMD, RL_EEMODE_OFF);
	}

	/* Enable transmit and receive.*/
	CSR_WRITE_1(sc, RL_COMMAND, RL_CMD_TX_ENB|RL_CMD_RX_ENB);

	/* Set the initial TX configuration.*/
	CSR_WRITE_4(sc, RL_TXCFG, RL_TXCFG_CONFIG);

	/* Set the initial RX configuration.*/
	if (sc->mac_style == MAC_STYLE_1) {
		CSR_WRITE_4(sc, RL_RXCFG, rxcfg | RL_RXCFG_CONFIG_1);
		CSR_WRITE_1(sc, 0xec, 0x3f);
	} else if (sc->mac_style == MAC_STYLE_2) {
		CSR_WRITE_4(sc, RL_RXCFG, rxcfg | RL_RXCFG_CONFIG_2);
		CSR_WRITE_1(sc, 0xec, 0x3f);
	}

	if ((sc->rl_type == MACFG_3) || (sc->rl_type == MACFG_4)) {
		CSR_WRITE_2(sc, RL_RxMaxSize, 0x800);	/* RMS */
	}

	CSR_WRITE_1(sc, RL_CFG1, RL_CFG1_DRVLOAD|RL_CFG1_FULLDUPLEX);

#if OS_VER<VERSION(6,0)
	ifp->if_flags |= IFF_RUNNING;
	ifp->if_flags &= ~IFF_OACTIVE;
#else
	ifp->if_drv_flags |= IFF_DRV_RUNNING;
	ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
#endif

	/*(void)splx(s);*/
	RL_UNLOCK(sc);

	sc->rl_stat_ch = timeout(rl_tick, sc, hz);

	return;
}

/*
 * Initialize the transmit descriptors.
 */
static int rl_var_init(struct rl_softc *sc)
{
	int			i;
	struct rl_chain_data	*cd;
	union RxDesc *rxptr;
	union TxDesc *txptr;

	if(!sc->DESCRIPTOR_MODE)
	{
		CSR_WRITE_4(sc, RL_RXADDR, vtophys(sc->rl_cdata.rl_rx_buf));
		cd = &sc->rl_cdata;
		for (i = 0; i < RL_TX_LIST_CNT; i++) {
			cd->rl_tx_chain[i] = NULL;
			CSR_WRITE_4(sc,
			    RL_TXADDR0 + (i * sizeof(u_int32_t)), 0x0000000);
		}
		sc->rl_cdata.cur_tx = 0;
		sc->rl_cdata.last_tx = 0;
	}
	else
	{
		uint32_t	phyAddr[2];
		sc->rl_desc.rx_cur_index = 0;
		sc->rl_desc.rx_last_index = 0;
		rxptr = sc->rl_desc.rx_desc;
		for(i=0;i<RL_RX_BUF_NUM;i++)
		{
			memset(&rxptr[i], 0, sizeof(union RxDesc));
			rxptr[i].so0.OWN=1;
			if(i==(RL_RX_BUF_NUM-1))
				rxptr[i].so0.EOR=1;
			rxptr[i].so0.Frame_Length = RL_BUF_SIZE;
			*((uint64_t *)&rxptr[i].so0.RxBuffL) = htole64(vtophys(sc->rl_desc.rx_buf[i]->m_data));
		}
		/* set Rx descriptor address */
		*((uint64_t *)phyAddr) = htole64(vtophys(sc->rl_desc.rx_desc));
		CSR_WRITE_4(sc, 0xe4, phyAddr[0]);
		CSR_WRITE_4(sc, 0xe8, phyAddr[1]);

		sc->rl_desc.tx_cur_index=0;
		sc->rl_desc.tx_last_index=0;
		txptr=sc->rl_desc.tx_desc;
		for(i=0;i<RL_TX_BUF_NUM;i++)
		{
			memset(&txptr[i], 0, sizeof(union TxDesc));
			if(i==(RL_TX_BUF_NUM-1))
				txptr[i].so1.EOR=1;
		}
		/* set Tx descriptor address */
		*((uint64_t *)phyAddr) = htole64(vtophys(sc->rl_desc.tx_desc));
		CSR_WRITE_4(sc, 0x20, phyAddr[0]);
		CSR_WRITE_4(sc, 0x24, phyAddr[1]);
	}
	return(0);
}

static void rl_reset(struct rl_softc *sc)
{
	register int		i;

	switch(sc->rl_type)
	{
		case MACFG_16:
		case MACFG_17:
		case MACFG_18:
		case MACFG_31:
		case MACFG_32:
		case MACFG_33:
			CSR_WRITE_1(sc, RL_COMMAND, 0x8C);
			DELAY(100);
			break;
	}

	CSR_WRITE_1(sc, RL_COMMAND, RL_CMD_RESET);

	for (i = 0; i < RL_TIMEOUT; i++) {
		DELAY(10);
		if (!(CSR_READ_1(sc, RL_COMMAND) & RL_CMD_RESET))
			break;
	}

	if (i == RL_TIMEOUT)
		printf("rl%d: reset never completed!\n", sc->rl_unit);

        return;
}

/*
 * Stop the adapter and free any mbufs allocated to the
 * RX and TX lists.
 */
static void rl_stop(struct rl_softc *sc)		/* Stop Driver */
{
	register int		i;
	struct ifnet		*ifp;

/*	RL_LOCK_ASSERT(sc);*/

#if OS_VER<VERSION(6,0)
	ifp = &sc->arpcom.ac_if;
#else
	ifp = sc->rl_ifp;
#endif
	ifp->if_timer = 0;

	untimeout(rl_tick, sc, sc->rl_stat_ch);

	CSR_WRITE_4(sc, RL_RXCFG, 0x0000);
	DELAY(100);

	CSR_WRITE_1(sc, RL_COMMAND, 0x00);
	CSR_WRITE_2(sc, RL_IMR, 0x0000);

	/*
	 * Free the TX list buffers.
	 */
	if(!sc->DESCRIPTOR_MODE)
		for (i = 0; i < RL_TX_LIST_CNT; i++)
		{
			if (sc->rl_cdata.rl_tx_chain[i] != NULL) {
				m_freem(sc->rl_cdata.rl_tx_chain[i]);
				sc->rl_cdata.rl_tx_chain[i] = NULL;
				CSR_WRITE_4(sc, RL_TXADDR0 + i, 0x0000000);
			}
		}
	else
	{
		while(sc->rl_desc.tx_last_index!=sc->rl_desc.tx_cur_index)
		{
			if(sc->rl_desc.tx_buf[sc->rl_desc.tx_last_index]!=NULL)
			{
				m_freem(sc->rl_desc.tx_buf[sc->rl_desc.tx_last_index]);
				sc->rl_desc.tx_buf[sc->rl_desc.tx_last_index] = NULL;
			}
			sc->rl_desc.tx_last_index = (sc->rl_desc.tx_last_index+1)%RL_TX_BUF_NUM;
		}
		CSR_WRITE_2(sc, 0xe0, 0x0000);
	}

#if OS_VER<VERSION(6,0)
	ifp->if_flags &= ~(IFF_RUNNING | IFF_OACTIVE);
#else
	ifp->if_drv_flags &= ~(IFF_DRV_RUNNING | IFF_DRV_OACTIVE);
#endif

	return;
}

/*
 * Main transmit routine.
 */
static void rl_start(struct ifnet *ifp)	/* Transmit Packet*/
{
	struct rl_softc		*sc;
	struct mbuf		*m_head = NULL;

#if 0	/*print the destination and source MAC addresses of tx packets*/
	int i;
#endif

	sc = ifp->if_softc;	/* Paste to ifp in function rl_attach(dev) */

	RL_LOCK(sc);

/*	RL_LOCK_ASSERT(sc);*/

	if((sc->driver_detach == 1)||(sc->rx_fifo_overflow != 0)){
		RL_UNLOCK(sc);
		return;
	}

	if(!sc->DESCRIPTOR_MODE)
	{
		while(RL_CUR_TXMBUF(sc) == NULL) {	/*  RL_CUR_TXMBUF(x)==>(x->rl_cdata.rl_tx_chain[x->rl_cdata.cur_tx]) */
			IFQ_DRV_DEQUEUE(&ifp->if_snd, m_head);	/* Remove(get) data from system transmit queue */
			if (m_head == NULL)
				break;

			if (rl_encap(sc, m_head))
			{
				IFQ_DRV_PREPEND(&ifp->if_snd, m_head);	/*  Case encap fail */
#if OS_VER<VERSION(6,0)
				ifp->if_flags |= IFF_OACTIVE;
#else
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
#endif
				break;
			}

			/*
			 * If there's a BPF listener, bounce a copy of this frame
			 * to him.
			 */
			if (ifp->if_bpf)
/*#if OS_VER < VERSION(5,1)*/
#if OS_VER < VERSION(4,9)
				bpf_mtap(ifp, RL_CUR_TXMBUF(sc));
#else
				bpf_mtap(ifp->if_bpf, RL_CUR_TXMBUF(sc));
#endif

			/*
			 * Transmit the frame.
	 		 */
			CSR_WRITE_4(sc, RL_CUR_TXADDR(sc),		/* CSR_WRITE_4(sc, reg, val) */
			    vtophys(mtod(RL_CUR_TXMBUF(sc), caddr_t)));	/* RL_CUR_TXMBUF(x)==> (x->rl_cdata.rl_tx_chain[x->rl_cdata.cur_tx]) */
			CSR_WRITE_4(sc, RL_CUR_TXSTAT(sc),		/* mtod(m, t)==> ((t)((m)->m_data)) */
			    RL_TXREV(sc->rl_txrev) |
			    RL_CUR_TXMBUF(sc)->m_pkthdr.len);

			RL_INC(sc->rl_cdata.cur_tx);			/* Move current Tx index */
		}

		/*
		 * We broke out of the loop because all our TX slots are
		 * full. Mark the NIC as busy until it drains some of the
		 * packets from the queue.
		 */
		if (RL_CUR_TXMBUF(sc) != NULL)				/* Tx buffer chain full */
#if OS_VER<VERSION(6,0)
			ifp->if_flags |= IFF_OACTIVE;
#else
			ifp->if_drv_flags |= IFF_DRV_OACTIVE;
#endif

		/*
		 * Set a timeout in case the chip goes out to lunch.
		 */
		ifp->if_timer = 5;
	}
	else	/* Descriptor Mode (C+ mode)*/
	{
	/*	printf("T");*/

		while(1)
		{
			int fs=1,ls=0,TxLen=0,PktLen,limit;
			struct mbuf *ptr;
			IFQ_DRV_DEQUEUE(&ifp->if_snd, m_head);	/* Remove(get) data from system transmit queue */
			if (m_head == NULL){
				break;
			}

#if 0	/*print the destination and source MAC addresses of tx packets*/
	{
	unsigned char *tmp_p;
	tmp_p = (unsigned char*) m_head->m_data;
	for(i=0;i<32;i++){
		printf("%2.2X ",*(tmp_p+i));
		if((i % 16) == 15)
			printf("\n");
	}
	printf("\n");
	}
#endif


			if(CountMbufNum(m_head,&limit)>CountFreeTxDescNum(sc->rl_desc))	/* No enough descriptor */
			{
				IFQ_DRV_PREPEND(&ifp->if_snd, m_head);
#if OS_VER<VERSION(6,0)
				ifp->if_flags |= IFF_OACTIVE;
#else
				ifp->if_drv_flags |= IFF_DRV_OACTIVE;
#endif
				break;
			}
			if (ifp->if_bpf)			/* If there's a BPF listener, bounce a copy of this frame to him. */
/*#if OS_VER < VERSION(5,1)*/
#if OS_VER < VERSION(4,9)
				bpf_mtap(ifp,m_head);
#else
				bpf_mtap(ifp->if_bpf,m_head);
#endif
			if(limit)				/* At least one mbuf data size small than RL_MINI_DESC_SIZE */
			{
#ifdef _DEBUG_
				ptr=m_head;
				printf("Limit=%d",limit);
				while(ptr!=NULL)
				{
					printf(", len=%d T=%d F=%d",ptr->m_len,ptr->m_type,ptr->m_flags);
					ptr=ptr->m_next;
				}
				printf("\n");
				printf("===== Reach limit ======\n");
#endif
				if (rl_encap(sc, m_head))
				{
					IFQ_DRV_PREPEND(&ifp->if_snd, m_head);
#if OS_VER<VERSION(6,0)
					ifp->if_flags |= IFF_OACTIVE;
#else
					ifp->if_drv_flags |= IFF_DRV_OACTIVE;
#endif
					break;
				}
				m_head = sc->rl_desc.tx_buf[sc->rl_desc.tx_cur_index];
				WritePacket(sc,m_head->m_data,m_head->m_len,1,1);
				continue;
			}

			ptr = m_head;
			PktLen = ptr->m_pkthdr.len;
#ifdef _DEBUG_
			printf("PktLen=%d",PktLen);
#endif
			while(ptr!=NULL)
			{
				if(ptr->m_len >0)
				{
#ifdef _DEBUG_
					printf(", len=%d T=%d F=%d",ptr->m_len,ptr->m_type,ptr->m_flags);
#endif
					TxLen += ptr->m_len;
					if(TxLen >= PktLen)
					{
						ls=1;
						sc->rl_desc.tx_buf[sc->rl_desc.tx_cur_index] = m_head;
					}
					else
						sc->rl_desc.tx_buf[sc->rl_desc.tx_cur_index] = NULL;

					WritePacket(sc,ptr->m_data,ptr->m_len,fs,ls);
					fs=0;
				}
				ptr = ptr->m_next;
			}
#ifdef _DEBUG_
			printf("\n");
#endif
		}
		ifp->if_timer = 5;
	/*	printf("t");*/
	}

	RL_UNLOCK(sc);

	return;
}

/*
 * Encapsulate an mbuf chain in a descriptor by coupling the mbuf data
 * pointers to the fragment pointers.
 */
static int rl_encap(struct rl_softc *sc,struct mbuf *m_head)		/* Only used in ~C+ mode */
{
	struct mbuf		*m_new = NULL;

	m_new = m_defrag(m_head, M_DONTWAIT);

	if (m_new == NULL) {
		printf("rl%d: no memory for tx list", sc->rl_unit);
		return (1);
	}
	m_head = m_new;

	/* Pad frames to at least 60 bytes. */
	if (m_head->m_pkthdr.len < RL_MIN_FRAMELEN) {	/* Case length < 60 bytes */
		/*
		 * Make security concious people happy: zero out the
		 * bytes in the pad area, since we don't know what
		 * this mbuf cluster buffer's previous user might
		 * have left in it.
	 	 */
		bzero(mtod(m_head, char *) + m_head->m_pkthdr.len,
		     RL_MIN_FRAMELEN - m_head->m_pkthdr.len);
		m_head->m_pkthdr.len = RL_MIN_FRAMELEN;
		m_head->m_len = m_head->m_pkthdr.len;
	}
	if(!sc->DESCRIPTOR_MODE)
		RL_CUR_TXMBUF(sc) = m_head;		/* Save pointer of buffer to sc */
	else
		sc->rl_desc.tx_buf[sc->rl_desc.tx_cur_index] = m_head;

	return(0);
}

static void WritePacket(struct rl_softc	*sc, caddr_t addr, int len,int fs_flag,int ls_flag)
{
	union TxDesc *txptr;
	txptr=&(sc->rl_desc.tx_desc[sc->rl_desc.tx_cur_index]);

	txptr->ul[0]&=0x40000000;
	txptr->ul[1]=0;

	if(fs_flag)
		txptr->so1.FS=1;
	if(ls_flag)
		txptr->so1.LS=1;
	txptr->so1.Frame_Length = len;
	*((uint64_t *)&txptr->so1.TxBuffL) = htole64(vtophys(addr));
//	txptr->so1.TxBuffL=vtophys(addr);
	txptr->so1.OWN = 1;

	if (ls_flag) {
		if (sc->DESCRIPTOR_MODE == TRUE) {
			CSR_WRITE_1(sc, RL_TPPOLL, RL_NPQ);
		} else {
			CSR_WRITE_1(sc, 0xd9,0x40);
		}
	}

	sc->rl_desc.tx_cur_index = (sc->rl_desc.tx_cur_index+1)%RL_TX_BUF_NUM;
}

static int CountFreeTxDescNum(struct rl_descriptor desc)
{
	int ret=desc.tx_last_index-desc.tx_cur_index;
	if(ret<=0)
		ret+=RL_TX_BUF_NUM;
	ret--;
	return ret;
}

static int CountMbufNum(struct mbuf *m_head,int *limit)
{
	int ret=0;
	struct mbuf *ptr=m_head;

	*limit=0;	/* 0:no limit find, 1:intermediate mbuf data size < RL_MINI_DESC_SIZE byte */
	while(ptr!=NULL)
	{
		if(ptr->m_len >0)
		{
			ret++;
			if(ptr->m_len<RL_MINI_DESC_SIZE && ptr->m_next!=NULL)	/* except last descriptor */
				*limit=1;
		}
		ptr=ptr->m_next;
	}
	return ret;
}

/*
 * A frame was downloaded to the chip. It's safe for us to clean up
 * the list buffers.
 */
static void rl_txeof(sc)	/* Transmit OK/ERR handler */
	struct rl_softc		*sc;
{
	if(!sc->DESCRIPTOR_MODE)
	{
		struct ifnet		*ifp;
		u_int32_t		txstat;

#if OS_VER<VERSION(6,0)
		ifp = &sc->arpcom.ac_if;
#else
		ifp = sc->rl_ifp;
#endif

		/* Clear the timeout timer. */
		ifp->if_timer = 0;

		/*
		 * Go through our tx list and free mbufs for those
		 * frames that have been uploaded.
		 */
		do {
			txstat = CSR_READ_4(sc, RL_LAST_TXSTAT(sc));
			if (!(txstat & (RL_TXSTAT_TX_OK|
			    RL_TXSTAT_TX_UNDERRUN|RL_TXSTAT_TXABRT)))
				break;

			ifp->if_collisions += (txstat & RL_TXSTAT_COLLCNT) >> 24;	/*  Add collision counter */

			if (RL_LAST_TXMBUF(sc) != NULL) {	/* Free Buffer */
				m_freem(RL_LAST_TXMBUF(sc));	/* RL_LAST_TXMBUF(x)==>(x->rl_cdata.rl_tx_chain[x->rl_cdata.last_tx]) */
				RL_LAST_TXMBUF(sc) = NULL;
			}
			if (txstat & RL_TXSTAT_TX_OK)		/* Case Tx OK */
				ifp->if_opackets++;
			else {					/* Case Tx ERR */
				int			oldrev;
				ifp->if_oerrors++;
				if ((txstat & RL_TXSTAT_TXABRT) ||
				    (txstat & RL_TXSTAT_OUTOFWIN))
					CSR_WRITE_4(sc, RL_TXCFG, RL_TXCFG_CONFIG);
				oldrev = sc->rl_txrev;
				/* error recovery */
				rl_reset(sc);
				rl_init(sc);

				if (txstat & RL_TXSTAT_TX_UNDERRUN)
					sc->rl_txrev = oldrev + 32;
				return;
			}
			RL_INC(sc->rl_cdata.last_tx);
#if OS_VER<VERSION(6,0)
			ifp->if_flags &= ~IFF_OACTIVE;
#else
			ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
#endif
		} while (sc->rl_cdata.last_tx != sc->rl_cdata.cur_tx);
	}
	else	/* Descriptor Mode */
	{
		union TxDesc *txptr;
		struct ifnet		*ifp;

	/*	printf("X");*/

#if OS_VER<VERSION(6,0)
		ifp = &sc->arpcom.ac_if;
#else
		ifp = sc->rl_ifp;
#endif

		/* Clear the timeout timer. */
		ifp->if_timer = 0;

		txptr=&(sc->rl_desc.tx_desc[sc->rl_desc.tx_last_index]);
		while(txptr->so1.OWN==0 && sc->rl_desc.tx_last_index!=sc->rl_desc.tx_cur_index)
		{
#ifdef _DEBUG_
			printf("**** Tx OK  ****\n");
#endif
			if(sc->rl_desc.tx_buf[sc->rl_desc.tx_last_index]!=NULL)
			{
				m_freem(sc->rl_desc.tx_buf[sc->rl_desc.tx_last_index]);	/* Free Current MBuf in a Mbuf list*/
				sc->rl_desc.tx_buf[sc->rl_desc.tx_last_index] = NULL;
			}

			sc->rl_desc.tx_last_index = (sc->rl_desc.tx_last_index+1)%RL_TX_BUF_NUM;
			txptr=&sc->rl_desc.tx_desc[sc->rl_desc.tx_last_index];
			ifp->if_opackets++;
#if OS_VER<VERSION(6,0)
			ifp->if_flags &= ~IFF_OACTIVE;
#else
			ifp->if_drv_flags &= ~IFF_DRV_OACTIVE;
#endif
		}
	}
	return;
}

/*
 * A frame has been uploaded: pass the resulting mbuf chain up to
 * the higher level protocols.
 *
 * You know there's something wrong with a PCI bus-master chip design
 * when you have to use m_devget().
 *
 * The receive operation is badly documented in the datasheet, so I'll
 * attempt to document it here. The driver provides a buffer area and
 * places its base address in the RX buffer start address register.
 * The chip then begins copying frames into the RX buffer. Each frame
 * is preceeded by a 32-bit RX status word which specifies the length
 * of the frame and certain other status bits. Each frame (starting with
 * the status word) is also 32-bit aligned. The frame length is in the
 * first 16 bits of the status word; the lower 15 bits correspond with
 * the 'rx status register' mentioned in the datasheet.
 *
 * Note: to make the Alpha happy, the frame payload needs to be aligned
 * on a 32-bit boundary. To achieve this, we cheat a bit by copying from
 * the ring buffer starting at an address two bytes before the actual
 * data location. We can then shave off the first two bytes using m_adj().
 * The reason we do this is because m_devget() doesn't let us specify an
 * offset into the mbuf storage space, so we have to artificially create
 * one. The ring is allocated in such a way that there are a few unused
 * bytes of space preceecing it so that it will be safe for us to do the
 * 2-byte backstep even if reading from the ring at offset 0.
 */
static void rl_rxeof(sc)	/* Receive Data OK/ERR handler */
	struct rl_softc		*sc;
{
	if(!sc->DESCRIPTOR_MODE)
	{
#if OS_VER < VERSION(4,9)
	        struct ether_header	*eh;
#endif
        	struct mbuf		*m;
	        struct ifnet		*ifp;
		int			total_len = 0;
		u_int32_t		rxstat;
		caddr_t			rxbufpos;
		int			wrap = 0;
		u_int16_t		cur_rx;
		u_int16_t		limit;
		u_int16_t		rx_bytes = 0, max_bytes;

#if OS_VER<VERSION(6,0)
		ifp = &sc->arpcom.ac_if;
#else
		ifp = sc->rl_ifp;
#endif

		cur_rx = (CSR_READ_2(sc, RL_CURRXADDR) + 16) % RL_RXBUFLEN;

		/* Do not try to read past this point. */
		limit = CSR_READ_2(sc, RL_CURRXBUF) % RL_RXBUFLEN;

		if (limit < cur_rx)
			max_bytes = (RL_RXBUFLEN - cur_rx) + limit;
		else
			max_bytes = limit - cur_rx;

		while((CSR_READ_1(sc, RL_COMMAND) & RL_CMD_EMPTY_RXBUF) == 0) {
			rxbufpos = sc->rl_cdata.rl_rx_buf + cur_rx;
			rxstat = *(u_int32_t *)rxbufpos;

			/*
			 * Here's a totally undocumented fact for you. When the
			 * RealTek chip is in the process of copying a packet into
			 * RAM for you, the length will be 0xfff0. If you spot a
			 * packet header with this value, you need to stop. The
			 * datasheet makes absolutely no mention of this and
			 * RealTek should be shot for this.
			 */
			if ((u_int16_t)(rxstat >> 16) == RL_RXSTAT_UNFINISHED)
				break;

			if (!(rxstat & RL_RXSTAT_RXOK)) {
				ifp->if_ierrors++;
				rl_init(sc);
				return;
			}

			/* No errors; receive the packet. */
			total_len = rxstat >> 16;
			rx_bytes += total_len + 4;

			/*
			 * XXX The RealTek chip includes the CRC with every
			 * received frame, and there's no way to turn this
			 * behavior off (at least, I can't find anything in
		 	 * the manual that explains how to do it) so we have
			 * to trim off the CRC manually.
			 */
			total_len -= ETHER_CRC_LEN;

			/*
			 * Avoid trying to read more bytes than we know
			 * the chip has prepared for us.
			 */
			if (rx_bytes > max_bytes)
				break;

			rxbufpos = sc->rl_cdata.rl_rx_buf +
				((cur_rx + sizeof(u_int32_t)) % RL_RXBUFLEN);

			if (rxbufpos == (sc->rl_cdata.rl_rx_buf + RL_RXBUFLEN))
				rxbufpos = sc->rl_cdata.rl_rx_buf;

			wrap = (sc->rl_cdata.rl_rx_buf + RL_RXBUFLEN) - rxbufpos;

			if (total_len > wrap) {
				/*
				 * Fool m_devget() into thinking we want to copy
				 * the whole buffer so we don't end up fragmenting
				 * the data.
				 */
				m = m_devget(rxbufpos - RL_ETHER_ALIGN,
				    total_len + RL_ETHER_ALIGN, 0, ifp, NULL);
				if (m == NULL) {
					ifp->if_ierrors++;
					printf("rl%d: out of mbufs, tried to "
					    "copy %d bytes\n", sc->rl_unit, wrap);
				} else {
					m_adj(m, RL_ETHER_ALIGN);
					m_copyback(m, wrap, total_len - wrap,
						sc->rl_cdata.rl_rx_buf);
				}
				cur_rx = (total_len - wrap + ETHER_CRC_LEN);
			} else {
				m = m_devget(rxbufpos - RL_ETHER_ALIGN,
				    total_len + RL_ETHER_ALIGN, 0, ifp, NULL);
				if (m == NULL) {
					ifp->if_ierrors++;
					printf("rl%d: out of mbufs, tried to "
					    "copy %d bytes\n", sc->rl_unit, total_len);
				} else
					m_adj(m, RL_ETHER_ALIGN);
				cur_rx += total_len + 4 + ETHER_CRC_LEN;
			}

			/*
			 * Round up to 32-bit boundary.
			 */
			cur_rx = (cur_rx + 3) & ~3;
			CSR_WRITE_2(sc, RL_CURRXADDR, cur_rx - 16);

			if (m == NULL)
				continue;

			ifp->if_ipackets++;

/*#if OS_VER < VERSION(5,1)*/

			RL_UNLOCK(sc);
#if OS_VER < VERSION(4,9)
			/* Remove header from mbuf and pass it on. */
			eh = mtod(m, struct ether_header *);
			m_adj(m, sizeof(struct ether_header));
			ether_input(ifp, eh, m);
#else
			(*ifp->if_input)(ifp, m);
#endif
			RL_LOCK(sc);

		}
	}
	else	/* Descriptor Mode */
	{
		struct ether_header	*eh;
		struct mbuf		*m;
		struct ifnet		*ifp;
		union RxDesc *rxptr;
		int bError;
		struct mbuf *buf;

/*		RL_LOCK_ASSERT(sc);*/

#if OS_VER<VERSION(6,0)
		ifp = &sc->arpcom.ac_if;
#else
		ifp = sc->rl_ifp;
#endif
		rxptr=&(sc->rl_desc.rx_desc[sc->rl_desc.rx_cur_index]);
		while(rxptr->so0.OWN==0)	/* Receive OK */
		{
			bError = 0;

			/* Check if this packet is received correctly*/
			if(sc->rl_type==RL_8139)
			{	/*8139C+*/
				if(rxptr->ul[0]&0x100000)	/* Check RES bit*/
				{
					bError=1;
					goto update_desc;
				}
			}
			else if(rxptr->ul[0]&0x200000)	/*Check RES bit*/
			{
				bError=1;
				goto update_desc;
			}

			buf = m_getcl(M_DONTWAIT, MT_DATA, M_PKTHDR); /* Alloc a new mbuf */
			if(buf==NULL)
			{
				bError=1;
				goto update_desc;
			}

			m = sc->rl_desc.rx_buf[sc->rl_desc.rx_cur_index];
			sc->rl_desc.rx_buf[sc->rl_desc.rx_cur_index] = buf;
			m->m_pkthdr.len = m->m_len = rxptr->so0.Frame_Length-ETHER_CRC_LEN;
			m->m_pkthdr.rcvif = ifp;

			eh = mtod(m, struct ether_header *);
			ifp->if_ipackets++;
#ifdef _DEBUG_
			printf("Rcv Packet, Len=%d \n",m->m_len);
#endif

			RL_UNLOCK(sc);

/*#if OS_VER < VERSION(5,1)*/
#if OS_VER < VERSION(4,9)
			/* Remove header from mbuf and pass it on. */
			m_adj(m, sizeof(struct ether_header));
			ether_input(ifp, eh, m);
#else
			(*ifp->if_input)(ifp, m);
#endif
			RL_LOCK(sc);

update_desc:
			rxptr->ul[0]&=0x40000000;	/* keep EOR bit */
			rxptr->ul[1]=0;

			rxptr->so0.Frame_Length = RL_BUF_SIZE;
			if(!bError)
			{
				*((uint64_t *)&rxptr->so0.RxBuffL) = htole64(vtophys(sc->rl_desc.rx_buf[sc->rl_desc.rx_cur_index]->m_data));
//				rxptr->so0.RxBuffL=vtophys(sc->rl_desc.rx_buf[sc->rl_desc.rx_cur_index]->m_data);
			}
			rxptr->so0.OWN=1;
			sc->rl_desc.rx_cur_index = (sc->rl_desc.rx_cur_index+1)%RL_RX_BUF_NUM;
			rxptr=&sc->rl_desc.rx_desc[sc->rl_desc.rx_cur_index];
		}
	}
	return;
}

static void rl_intr(arg)	/* Interrupt Handler */
	void			*arg;
{
	struct rl_softc		*sc;
	struct ifnet		*ifp;
	u_int16_t		status;
	u_int8_t		link_status;

	int	intr_loop=INTR_MAX_LOOP;
	/*int	s;*/
	sc = arg;

	RL_LOCK(sc);

#if OS_VER<VERSION(6,0)
	ifp = &sc->arpcom.ac_if;
#else
	ifp = sc->rl_ifp;
#endif

	/* Disable interrupts. */
	CSR_WRITE_2(sc, RL_IMR, 0x0000);

	for (intr_loop=INTR_MAX_LOOP;intr_loop>0;intr_loop--) {

		status = CSR_READ_2(sc, RL_ISR);

		if (status){
			CSR_WRITE_2(sc, RL_ISR, status & 0xffbf);
		}

		if ((status & RL_INTRS) == 0){
			break;
		}

		if ((status & RL_ISR_RX_OK) || (status & RL_ISR_RX_ERR) ||(status & RL_ISR_FIFO_OFLOW) || (status & RL_ISR_RX_OVERRUN)){
			rl_rxeof(sc);
		}

		if (sc->rl_type == MACFG_21){
			if (status & RL_ISR_FIFO_OFLOW){
				sc->rx_fifo_overflow = 1;
				CSR_WRITE_2(sc, 0x00e2, 0x0000);
				CSR_WRITE_4(sc, 0x0048, 0x4000);
				CSR_WRITE_4(sc, 0x0058, 0x4000);
			}else{
				sc->rx_fifo_overflow = 0;
				CSR_WRITE_4(sc,RL_CPCR,0x51512082);
			}

			if (status & RL_ISR_PCS_TIMEOUT){
				if((status & RL_ISR_FIFO_OFLOW) &&
				   (!(status & (RL_ISR_RX_OK | RL_ISR_TX_OK | RL_ISR_RX_OVERRUN)))){
					rl_reset(sc);
					rl_init(sc);
					sc->rx_fifo_overflow = 0;
					CSR_WRITE_2(sc, RL_ISR, RL_ISR_FIFO_OFLOW);
				}
			}
		}

		if(status & RL_ISR_LINKCHG){
			link_status = CSR_READ_1(sc, RL_PHY_STATUS);
			if(link_status & 0x02){
				ifp->if_link_state = LINK_STATE_UP;
				//PHY CONFIG
				MP_WritePhyUshort(sc, 0x1f, 0x0001);
				MP_WritePhyUshort(sc, 0x1f, 0xF41B);
				MP_WritePhyUshort(sc, 0x1f, 0x0000);
			}else{
				ifp->if_link_state = LINK_STATE_DOWN;
				//PHY CONFIG
				MP_WritePhyUshort(sc, 0x1f, 0x0001);
				MP_WritePhyUshort(sc, 0x1f, 0xF01B);
				MP_WritePhyUshort(sc, 0x1f, 0x0000);
			}
		}

		if ((status & RL_ISR_TX_OK) || (status & RL_ISR_TX_ERR)){
			rl_txeof(sc);
		}

		if (status & RL_ISR_SYSTEM_ERR) {
			rl_reset(sc);
			rl_init(sc);
		}

		if ((status & RL_ISR_TX_OK) && (status & RL_ISR_TDU)){
			CSR_WRITE_1(sc, 0x38,0x40);
		}

	}

	/* Re-enable interrupts. */
	CSR_WRITE_2(sc, RL_IMR, RL_INTRS);

	if (ifp->if_snd.ifq_head != NULL)		/* Data in Tx buffer waiting for transimission */
		rl_start(ifp);

	RL_UNLOCK(sc);

	return;
}

/*
 * Calculate CRC of a multicast group address, return the upper 6 bits.
 */
static u_int8_t rl_calchash(addr)
	caddr_t			addr;
{
	u_int32_t		crc, carry;
	int			i, j;
	u_int8_t		c;

	/* Compute CRC for the address value. */
	crc = 0xFFFFFFFF; /* initial value */

	for (i = 0; i < 6; i++) {
		c = *(addr + i);
		for (j = 0; j < 8; j++) {
			carry = ((crc & 0x80000000) ? 1 : 0) ^ (c & 0x01);
			crc <<= 1;
			c >>= 1;
			if (carry)
				crc = (crc ^ 0x04c11db6) | carry;
		}
	}

	/* return the filter bit position */
	return(crc >> 26);
}

/*
 * Program the 64-bit multicast hash filter.
 */
static void rl_setmulti(sc)
	struct rl_softc		*sc;
{
	struct ifnet		*ifp;
	int			h = 0;
	u_int32_t		hashes[2] = { 0, 0 };
	struct ifmultiaddr	*ifma;
	u_int32_t		rxfilt;
	int			mcnt = 0;

#if OS_VER<VERSION(6,0)
	ifp = &sc->arpcom.ac_if;
#else
	ifp = sc->rl_ifp;
#endif

	rxfilt = CSR_READ_4(sc, RL_RXCFG);

	if (ifp->if_flags & IFF_ALLMULTI || ifp->if_flags & IFF_PROMISC) {
		rxfilt |= RL_RXCFG_RX_MULTI;
		CSR_WRITE_4(sc, RL_RXCFG, rxfilt);
		CSR_WRITE_4(sc, RL_MAR0, 0xFFFFFFFF);
		CSR_WRITE_4(sc, RL_MAR4, 0xFFFFFFFF);
		return;
	}

	/* first, zot all the existing hash bits */
	CSR_WRITE_4(sc, RL_MAR0, 0);
	CSR_WRITE_4(sc, RL_MAR4, 0);

	/* now program new ones */
/*#if OS_VER < VERSION(5,1)*/
#if OS_VER < VERSION(4,9)
	for (ifma = ifp->if_multiaddrs.lh_first; ifma != NULL;
				ifma = ifma->ifma_link.le_next)
#else
	TAILQ_FOREACH(ifma,&ifp->if_multiaddrs,ifma_link)
#endif
	{
		if (ifma->ifma_addr->sa_family != AF_LINK)
			continue;
		h = rl_calchash(LLADDR((struct sockaddr_dl *)ifma->ifma_addr));
		if (h < 32)
			hashes[0] |= (1 << h);
		else
			hashes[1] |= (1 << (h - 32));
		mcnt++;
	}

	if (mcnt)
		rxfilt |= RL_RXCFG_RX_MULTI;
	else
		rxfilt &= ~RL_RXCFG_RX_MULTI;

	CSR_WRITE_4(sc, RL_RXCFG, rxfilt);
	CSR_WRITE_4(sc, RL_MAR0, hashes[0]);
	CSR_WRITE_4(sc, RL_MAR4, hashes[1]);

	return;
}

static int rl_ioctl(ifp, command, data)
	struct ifnet		*ifp;
	u_long			command;
	caddr_t			data;
{
	struct rl_softc		*sc = ifp->if_softc;
	struct ifreq		*ifr = (struct ifreq *) data;
	/*int			s;*/
	int			error = 0;

	/*s = splimp();*/

	switch(command) {
	case SIOCSIFADDR:
	case SIOCGIFADDR:
	case SIOCSIFMTU:
		error = ether_ioctl(ifp, command, data);
		break;
	case SIOCSIFFLAGS:
		RL_LOCK(sc);
		if (ifp->if_flags & IFF_UP) {
			rl_init(sc);
		} else {
#if OS_VER<VERSION(6,0)
			if (ifp->if_flags & IFF_RUNNING)
#else
			if (ifp->if_flags & IFF_DRV_RUNNING)
#endif
				rl_stop(sc);
		}
		error = 0;
		RL_UNLOCK(sc);
		break;
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		RL_LOCK(sc);
		rl_setmulti(sc);
		RL_UNLOCK(sc);
		error = 0;
		break;
	case SIOCGIFMEDIA:
	case SIOCSIFMEDIA:
		error = ifmedia_ioctl(ifp, ifr, &sc->media, command);
		break;
	default:
		error = EINVAL;
		break;
	}

	/*(void)splx(s);*/

	return(error);
}

static void rl_tick(xsc)
	void			*xsc;
{	/*called per second*/
	struct rl_softc		*sc;
	int			s;

	s = splimp();

	sc = xsc;
	/*mii = device_get_softc(sc->rl_miibus);

	mii_tick(mii);*/

	splx(s);

	if (sc->rl_type == MACFG_3 &&
	    sc->rl_8169_MacVersion != 0 &&
	    sc->rl_8169_PhyVersion == 0) {
		static int count = 0;

		if (CSR_READ_1 (sc, 0x6C) & 0x02)
			count = 0;
		else
			count++;
		if (count> 14) {
			MP_WritePhyUshort (sc, 0, 0x9000);
			count = 0;
		}
	}

	sc->rl_stat_ch = timeout (rl_tick, sc, hz);

	return;
}

#if OS_VER<VERSION(7,0)
static void rl_watchdog(ifp)
	struct ifnet		*ifp;
{
	struct rl_softc		*sc;

	sc = ifp->if_softc;

	printf("rl%d: watchdog timeout\n", sc->rl_unit);
	ifp->if_oerrors++;

	rl_txeof(sc);
	rl_rxeof(sc);
	rl_init(sc);

	return;
}
#endif

static void rl_set_speed_and_duplex(struct rl_softc *sc, int gbcr, int anar, int bmcr)
{
	MP_WritePhyUshort(sc, RL_PHY_ANAR, anar | 0x0001);
	MP_WritePhyUshort(sc, RL_PHY_GBCR, gbcr);
	MP_WritePhyUshort(sc, RL_PHY_BMCR, bmcr);
}

/*
 * Set media options.
 */
static int rl_ifmedia_upd(ifp)
	struct ifnet		*ifp;
{
	struct rl_softc	*sc = ifp->if_softc;
	struct ifmedia	*ifm = &sc->media;
	int bmcr;
	int anar;
	int gbcr;

	if (IFM_TYPE(ifm->ifm_media) != IFM_ETHER)
		return(EINVAL);

	switch (IFM_SUBTYPE(ifm->ifm_media)) {
	case IFM_AUTO:
		anar = RL_PHY_100_FULL |
		       RL_PHY_100_HALF |
		       RL_PHY_10_FULL |
		       RL_PHY_10_HALF;
		gbcr = RL_PHY_1000_FULL |
		       RL_PHY_1000_HALF;
		break;
	case IFM_1000_SX:
#if __FreeBSD_version < 500000
	case IFM_1000_TX:
#else
	case IFM_1000_T:
#endif
		anar = RL_PHY_100_FULL |
		       RL_PHY_100_HALF |
		       RL_PHY_10_FULL |
		       RL_PHY_10_HALF;
		gbcr = RL_PHY_1000_FULL |
		       RL_PHY_1000_HALF;
		break;
	case IFM_100_TX:
		gbcr = MP_ReadPhyUshort(sc, RL_PHY_GBCR) &
		       ~(RL_PHY_1000_FULL | RL_PHY_1000_HALF);
		if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX) {
			anar = RL_PHY_100_FULL |
			       RL_PHY_100_HALF |
			       RL_PHY_10_FULL |
			       RL_PHY_10_HALF;
		} else {
			anar = RL_PHY_100_HALF |
			       RL_PHY_10_FULL |
			       RL_PHY_10_HALF;
		}
		break;
	case IFM_10_T:
		gbcr = MP_ReadPhyUshort(sc, RL_PHY_GBCR) &
		       ~(RL_PHY_1000_FULL | RL_PHY_1000_HALF);
		if ((ifm->ifm_media & IFM_GMASK) == IFM_FDX) {
			anar = RL_PHY_10_FULL |
			       RL_PHY_10_HALF;
		} else {
			anar = RL_PHY_10_HALF;
		}

		if (sc->rl_type == MACFG_13) {
			MP_WritePhyUshort(sc, RL_PHY_BMCR, 0x8000);
		}

		break;
	default:
		printf("rl%d: Unsupported media type\n", sc->rl_unit);
		return(0);
	}

	bmcr = RL_PHY_NWAY_EN | RL_PHY_RESTART_NWAY;

	rl_set_speed_and_duplex(sc, gbcr, anar, bmcr);

	return(0);
}

/*
 * Report current media status.
 */
static void rl_ifmedia_sts(ifp, ifmr)
	struct ifnet		*ifp;
	struct ifmediareq	*ifmr;
{
	struct rl_softc		*sc;

	sc = ifp->if_softc;

	ifmr->ifm_status = IFM_AVALID;
	ifmr->ifm_active = IFM_ETHER;
	if(sc->rl_type==RL_8139)
	{
		unsigned char msr;
		unsigned short lpar;

		msr=CSR_READ_1(sc,0x58);
		lpar=CSR_READ_2(sc,0x68);

		if((msr&0x04)==0)
			ifmr->ifm_status|= IFM_ACTIVE;
		else
			return;

		if(lpar!=0)	/* support NWAY */
		{
			if(lpar&0x100)
				ifmr->ifm_active|= IFM_100_TX|IFM_FDX;
			else if(lpar&0x80)
				ifmr->ifm_active|= IFM_100_TX|IFM_HDX;
			else if(lpar&0x40)
				ifmr->ifm_active|= IFM_10_T|IFM_FDX;
			else if(lpar&0x20)
				ifmr->ifm_active|= IFM_10_T|IFM_HDX;
		}
		else
		{
			ifmr->ifm_active|= IFM_HDX;
			if(msr&0x08)
				ifmr->ifm_active|= IFM_10_T;
			else
				ifmr->ifm_active|= IFM_100_TX;
		}
	} else {
		unsigned char msr;
		msr = CSR_READ_1(sc, 0x6c);

		if(msr & 0x02)
			ifmr->ifm_status |= IFM_ACTIVE;
		else
			return;

		if(msr & 0x01)
			ifmr->ifm_active |= IFM_FDX;
		else
			ifmr->ifm_active |= IFM_HDX;


		if(msr & 0x04)
			ifmr->ifm_active |= IFM_10_T;
		else if(msr & 0x08)
			ifmr->ifm_active |= IFM_100_TX;
		else if(msr & 0x10)
#if OS_VER < VERSION(4,9)
			ifmr->ifm_active |= IFM_1000_TX;
#else
			ifmr->ifm_active |= IFM_1000_T;
#endif
	}

	return;
}

static void rl_8169s_init(struct rl_softc *sc)
{
	int anar, gbcr, bmcr;
	u_int16_t Data;

	if(sc->rl_type==MACFG_3){
		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x06, 0x006e);
		MP_WritePhyUshort(sc, 0x08, 0x0708);
		MP_WritePhyUshort(sc, 0x15, 0x4000);
		MP_WritePhyUshort(sc, 0x18, 0x65c7);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x03, 0x00a1);
		MP_WritePhyUshort(sc, 0x02, 0x0008);
		MP_WritePhyUshort(sc, 0x01, 0x0120);
		MP_WritePhyUshort(sc, 0x00, 0x1000);
		MP_WritePhyUshort(sc, 0x04, 0x0800);
		MP_WritePhyUshort(sc, 0x04, 0x0000);

		MP_WritePhyUshort(sc, 0x03, 0xff41);
		MP_WritePhyUshort(sc, 0x02, 0xdf60);
		MP_WritePhyUshort(sc, 0x01, 0x0140);
		MP_WritePhyUshort(sc, 0x00, 0x0077);
		MP_WritePhyUshort(sc, 0x04, 0x7800);
		MP_WritePhyUshort(sc, 0x04, 0x7000);

		MP_WritePhyUshort(sc, 0x03, 0x802f);
		MP_WritePhyUshort(sc, 0x02, 0x4f02);
		MP_WritePhyUshort(sc, 0x01, 0x0409);
		MP_WritePhyUshort(sc, 0x00, 0xf0f9);
		MP_WritePhyUshort(sc, 0x04, 0x9800);
		MP_WritePhyUshort(sc, 0x04, 0x9000);

		MP_WritePhyUshort(sc, 0x03, 0xdf01);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0xff95);
		MP_WritePhyUshort(sc, 0x00, 0xba00);
		MP_WritePhyUshort(sc, 0x04, 0xa800);
		MP_WritePhyUshort(sc, 0x04, 0xa000);

		MP_WritePhyUshort(sc, 0x03, 0xff41);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0x0140);
		MP_WritePhyUshort(sc, 0x00, 0x00bb);
		MP_WritePhyUshort(sc, 0x04, 0xb800);
		MP_WritePhyUshort(sc, 0x04, 0xb000);

		MP_WritePhyUshort(sc, 0x03, 0xdf41);
		MP_WritePhyUshort(sc, 0x02, 0xdc60);
		MP_WritePhyUshort(sc, 0x01, 0x6340);
		MP_WritePhyUshort(sc, 0x00, 0x007d);
		MP_WritePhyUshort(sc, 0x04, 0xd800);
		MP_WritePhyUshort(sc, 0x04, 0xd000);

		MP_WritePhyUshort(sc, 0x03, 0xdf01);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0x100a);
		MP_WritePhyUshort(sc, 0x00, 0xa0ff);
		MP_WritePhyUshort(sc, 0x04, 0xf800);
		MP_WritePhyUshort(sc, 0x04, 0xf000);

		MP_WritePhyUshort(sc, 0x1f, 0x0000);
		MP_WritePhyUshort(sc, 0x0b, 0x0000);
		MP_WritePhyUshort(sc, 0x00, 0x9200);

		CSR_WRITE_1(sc, 0x82, 0x0d);
	} else if(sc->rl_type==MACFG_4) {
		MP_WritePhyUshort(sc, 0x1f, 0x0002);
		MP_WritePhyUshort(sc, 0x01, 0x90D0);
	} else if(sc->rl_type==MACFG_5) {
		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x04, 0x0000);
		MP_WritePhyUshort(sc, 0x03, 0x00a1);
		MP_WritePhyUshort(sc, 0x02, 0x0008);
		MP_WritePhyUshort(sc, 0x01, 0x0120);
		MP_WritePhyUshort(sc, 0x00, 0x1000);
		MP_WritePhyUshort(sc, 0x04, 0x0800);

		MP_WritePhyUshort(sc, 0x04, 0x9000);
		MP_WritePhyUshort(sc, 0x03, 0x820f);
		MP_WritePhyUshort(sc, 0x02, 0x4f02);
		MP_WritePhyUshort(sc, 0x01, 0x0409);
		MP_WritePhyUshort(sc, 0x00, 0xf099);
		MP_WritePhyUshort(sc, 0x04, 0x9800);

		MP_WritePhyUshort(sc, 0x04, 0xa000);
		MP_WritePhyUshort(sc, 0x03, 0xdf01);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0xff95);
		MP_WritePhyUshort(sc, 0x00, 0xba00);
		MP_WritePhyUshort(sc, 0x04, 0xa800);

		MP_WritePhyUshort(sc, 0x04, 0xf000);
		MP_WritePhyUshort(sc, 0x03, 0xdf01);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0x101a);
		MP_WritePhyUshort(sc, 0x00, 0xa0ff);
		MP_WritePhyUshort(sc, 0x04, 0xf800);
		MP_WritePhyUshort(sc, 0x04, 0x0000);
		MP_WritePhyUshort(sc, 0x1f, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x10, 0xf41b);
		MP_WritePhyUshort(sc, 0x14, 0xfb54);
		MP_WritePhyUshort(sc, 0x18, 0xf5c7);
		MP_WritePhyUshort(sc, 0x1f, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);
		MP_WritePhyUshort(sc, 0x1f, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x10, 0xf01b);

	} else if(sc->rl_type==MACFG_6) {
		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x04, 0x0000);
		MP_WritePhyUshort(sc, 0x03, 0x00a1);
		MP_WritePhyUshort(sc, 0x02, 0x0008);
		MP_WritePhyUshort(sc, 0x01, 0x0120);
		MP_WritePhyUshort(sc, 0x00, 0x1000);
		MP_WritePhyUshort(sc, 0x04, 0x0800);

		MP_WritePhyUshort(sc, 0x04, 0x9000);
		MP_WritePhyUshort(sc, 0x03, 0x820f);
		MP_WritePhyUshort(sc, 0x02, 0x4f02);
		MP_WritePhyUshort(sc, 0x01, 0x0409);
		MP_WritePhyUshort(sc, 0x00, 0xf099);
		MP_WritePhyUshort(sc, 0x04, 0x9800);

		MP_WritePhyUshort(sc, 0x04, 0xa000);
		MP_WritePhyUshort(sc, 0x03, 0xdf01);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0xff95);
		MP_WritePhyUshort(sc, 0x00, 0xba00);
		MP_WritePhyUshort(sc, 0x04, 0xa800);

		MP_WritePhyUshort(sc, 0x04, 0xf000);
		MP_WritePhyUshort(sc, 0x03, 0xdf01);
		MP_WritePhyUshort(sc, 0x02, 0xdf20);
		MP_WritePhyUshort(sc, 0x01, 0x101a);
		MP_WritePhyUshort(sc, 0x00, 0xa0ff);
		MP_WritePhyUshort(sc, 0x04, 0xf800);
		MP_WritePhyUshort(sc, 0x04, 0x0000);
		MP_WritePhyUshort(sc, 0x1f, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x0b, 0x8480);
		MP_WritePhyUshort(sc, 0x1f, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x18, 0x67c7);
		MP_WritePhyUshort(sc, 0x04, 0x2000);
		MP_WritePhyUshort(sc, 0x03, 0x002f);
		MP_WritePhyUshort(sc, 0x02, 0x4360);
		MP_WritePhyUshort(sc, 0x01, 0x0109);
		MP_WritePhyUshort(sc, 0x00, 0x3022);
		MP_WritePhyUshort(sc, 0x04, 0x2800);
		MP_WritePhyUshort(sc, 0x1f, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);
	} else if (sc->rl_type == MACFG_14) {
		MP_WritePhyUshort(sc, 0x1f, 0x0000);
		MP_WritePhyUshort(sc, 0x11, 0x15C0);
		MP_WritePhyUshort(sc, 0x19, 0x2080);

		MP_WritePhyUshort(sc, 0x1f, 0x0003);
		MP_WritePhyUshort(sc, 0x08, 0x441D);
		MP_WritePhyUshort(sc, 0x01, 0x9100);

	} else if (sc->rl_type == MACFG_15) {
//		MP_WritePhyUshort(sc, 0x1f, 0x0000);
//		MP_WritePhyUshort(sc, 0x11, 0x15C0);
//		MP_WritePhyUshort(sc, 0x19, 0x2080);

		MP_WritePhyUshort(sc, 0x1f, 0x0003);
		MP_WritePhyUshort(sc, 0x08, 0x441D);
		MP_WritePhyUshort(sc, 0x01, 0x9100);

		MP_WritePhyUshort(sc, 0x1f, 0x0000);
		Data = MP_ReadPhyUshort(sc, 0x10);
		Data |= 0x8000;
		MP_WritePhyUshort(sc, 0x10, Data);

	} else if (sc->rl_type == MACFG_16) {
		MP_WritePhyUshort(sc, 0x1f, 0x0000);
		Data = MP_ReadPhyUshort(sc, 0x10);
		Data |= 0x8000;
		MP_WritePhyUshort(sc, 0x10, Data);

	} else if (sc->rl_type == MACFG_24) {
		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x12, 0x2300);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x00, 0x88DE);
		MP_WritePhyUshort(sc, 0x01, 0x82B1);
		MP_WritePhyUshort(sc, 0x03, 0x7002);
		MP_WritePhyUshort(sc, 0x08, 0x9E30);
		MP_WritePhyUshort(sc, 0x09, 0x01F0);
		MP_WritePhyUshort(sc, 0x0A, 0x5500);
		MP_WritePhyUshort(sc, 0x0C, 0x00C8);

		MP_WritePhyUshort(sc, 0x1F, 0x0003);
		MP_WritePhyUshort(sc, 0x12, 0xC096);
		MP_WritePhyUshort(sc, 0x16, 0x000A);

		MP_WritePhyUshort(sc, 0x1F, 0x0000);
		MP_WritePhyUshort(sc, 0x14, MP_ReadPhyUshort(sc, 0x14) | (1 << 5));
		MP_WritePhyUshort(sc, 0x0d, MP_ReadPhyUshort(sc, 0x0d) | (1 << 5));

	} else if (sc->rl_type == MACFG_25) {
		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x12, 0x2300);
		MP_WritePhyUshort(sc, 0x03, 0x802f);
		MP_WritePhyUshort(sc, 0x02, 0x4f02);
		MP_WritePhyUshort(sc, 0x01, 0x0409);
		MP_WritePhyUshort(sc, 0x00, 0xf099);
		MP_WritePhyUshort(sc, 0x04, 0x9800);
		MP_WritePhyUshort(sc, 0x04, 0x9000);
		MP_WritePhyUshort(sc, 0x1D, 0x3D98);

		MP_WritePhyUshort(sc, 0x1f, 0x0002);
		MP_WritePhyUshort(sc, 0x00, 0x88DE);
		MP_WritePhyUshort(sc, 0x01, 0x82B1);
		MP_WritePhyUshort(sc, 0x0c, 0x7eb8);
		MP_WritePhyUshort(sc, 0x06, 0x0761);

		MP_WritePhyUshort(sc, 0x1f, 0x0003);
		MP_WritePhyUshort(sc, 0x16, 0x0f0a);

		MP_WritePhyUshort(sc, 0x1F, 0x0000);
		MP_WritePhyUshort(sc, 0x16, MP_ReadPhyUshort(sc, 0x16) | (1 << 0));
		MP_WritePhyUshort(sc, 0x14, MP_ReadPhyUshort(sc, 0x14) | (1 << 5));
		MP_WritePhyUshort(sc, 0x0d, MP_ReadPhyUshort(sc, 0x0d) | (1 << 5));

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);

	} else if (sc->rl_type == MACFG_27) {
		MP_WritePhyUshort(sc, 0x1F, 0x0000);
		MP_WritePhyUshort(sc, 0x0d, MP_ReadPhyUshort(sc, 0x0d) | (1 << 5));
		MP_WritePhyUshort(sc, 0x1F, 0x0000);

		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x1D, 0x3D98);
		MP_WritePhyUshort(sc, 0x1F, 0x0000);

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);

	} else if (sc->rl_type == MACFG_28) {
		MP_WritePhyUshort(sc, 0x1F, 0x0000);
		MP_WritePhyUshort(sc, 0x0d, MP_ReadPhyUshort(sc, 0x0d) | (1 << 5));

		MP_WritePhyUshort(sc, 0x1f, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);

	} else if (sc->rl_type == MACFG_26) {
		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x12, 0x2300);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);
		MP_WritePhyUshort(sc, 0x1D, 0x3D98);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x00, 0x88DE);
		MP_WritePhyUshort(sc, 0x01, 0x82B1);
		MP_WritePhyUshort(sc, 0x06, 0x5461);
		MP_WritePhyUshort(sc, 0x0C, 0x7EB8);

		MP_WritePhyUshort(sc, 0x1F, 0x0003);
		MP_WritePhyUshort(sc, 0x16, 0x0F0A);

		MP_WritePhyUshort(sc, 0x1F, 0x0000);
		MP_WritePhyUshort(sc, 0x16, MP_ReadPhyUshort(sc, 0x16) | (1 << 0));
		MP_WritePhyUshort(sc, 0x14, MP_ReadPhyUshort(sc, 0x14) | (1 << 5));
		MP_WritePhyUshort(sc, 0x0d, MP_ReadPhyUshort(sc, 0x0d) | (1 << 5));

	} else if (sc->rl_type == MACFG_31) {
		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x06, 0x4064);
		MP_WritePhyUshort(sc, 0x07, 0x2863);
		MP_WritePhyUshort(sc, 0x08, 0x059C);
		MP_WritePhyUshort(sc, 0x09, 0x26B4);
		MP_WritePhyUshort(sc, 0x0A, 0x6A19);
		MP_WritePhyUshort(sc, 0x0B, 0xBCC0);
		MP_WritePhyUshort(sc, 0x10, 0xF06D);
		MP_WritePhyUshort(sc, 0x14, 0x7F68);
		MP_WritePhyUshort(sc, 0x18, 0x7FD9);
		MP_WritePhyUshort(sc, 0x1C, 0xF0FF);
		MP_WritePhyUshort(sc, 0x1D, 0x3D9C);
		MP_WritePhyUshort(sc, 0x1F, 0x0003);
		MP_WritePhyUshort(sc, 0x12, 0xF49F);
		MP_WritePhyUshort(sc, 0x13, 0x070B);
		MP_WritePhyUshort(sc, 0x1A, 0x05AD);
		MP_WritePhyUshort(sc, 0x14, 0x94C0);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x0B) & 0xFF00;
		Data |= 0x0010;
		MP_WritePhyUshort(sc, 0x0B, Data);
		Data = MP_ReadPhyUshort(sc, 0x0C) & 0x00FF;
		Data |= 0xA200;
		MP_WritePhyUshort(sc, 0x0C, Data);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x06, 0x5561);
		MP_WritePhyUshort(sc, 0x1f, 0x0005);
		MP_WritePhyUshort(sc, 0x05, 0x8332);
		MP_WritePhyUshort(sc, 0x06, 0x5561);

		if(MP_ReadEfuse(sc, 0x01)==0xB1)
		{
			MP_WritePhyUshort(sc, 0x1f, 0x0002);
			MP_WritePhyUshort(sc, 0x05, 0x669a);
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0x8330);
			MP_WritePhyUshort(sc, 0x06, 0x669a);
			MP_WritePhyUshort(sc, 0x1f, 0x0002);
			Data = MP_ReadPhyUshort(sc, 0x0D);
			if((Data & 0x00FF)!=0x006C)
			{
				Data &= 0xFF00;
				MP_WritePhyUshort(sc, 0x0D, 0x0065|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0066|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0067|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0068|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0069|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x006A|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x006B|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x006C|Data);
			}
		}
		else
		{
			MP_WritePhyUshort(sc, 0x1f, 0x0002);
			MP_WritePhyUshort(sc, 0x05, 0x6662);
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0x8330);
			MP_WritePhyUshort(sc, 0x06, 0x6662);
		}

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x0D);
		Data |= 0x0300;
		MP_WritePhyUshort(sc, 0x0D, Data);
		Data = MP_ReadPhyUshort(sc, 0x0F);
		Data |= 0x0010;
		MP_WritePhyUshort(sc, 0x0F, Data);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x05);
		Data &= 0xF0FF;
		if(CSR_READ_1(sc, RL_PHY_STATUS)&0x10)
		{
			switch(Data)
			{
				case 0x609A:
					MP_WritePhyUshort(sc, 0x15, 0x3928);
					MP_WritePhyUshort(sc, 0x16, 0x794c);
					MP_WritePhyUshort(sc, 0x17, 0x79b4);
					break;
				case 0x6062:
					MP_WritePhyUshort(sc, 0x15, 0x3628);
					MP_WritePhyUshort(sc, 0x16, 0x764c);
					MP_WritePhyUshort(sc, 0x17, 0x76b4);
					break;
				case 0x2042:
					MP_WritePhyUshort(sc, 0x15, 0x3428);
					MP_WritePhyUshort(sc, 0x16, 0x344c);
					MP_WritePhyUshort(sc, 0x17, 0x34b4);
					break;
			}
			MP_WritePhyUshort(sc, 0x1F, 0x0003);
			MP_WritePhyUshort(sc, 0x17, 0x7bc0);
			MP_WritePhyUshort(sc, 0x18, 0xff46);
			MP_WritePhyUshort(sc, 0x19, 0x3229);
		}
		else
		{
			switch(Data)
			{
				case 0x609A:
					MP_WritePhyUshort(sc, 0x15, 0x3900);
					MP_WritePhyUshort(sc, 0x16, 0x7924);
					MP_WritePhyUshort(sc, 0x17, 0x7948);
					break;
				case 0x6062:
					MP_WritePhyUshort(sc, 0x15, 0x3600);
					MP_WritePhyUshort(sc, 0x16, 0x7624);
					MP_WritePhyUshort(sc, 0x17, 0x7648);
					break;
				case 0x2042:
					MP_WritePhyUshort(sc, 0x15, 0x3400);
					MP_WritePhyUshort(sc, 0x16, 0x3424);
					MP_WritePhyUshort(sc, 0x17, 0x3448);
					break;
			}
			MP_WritePhyUshort(sc, 0x1f, 0x0003);
			MP_WritePhyUshort(sc, 0x10, 0xAC7C);
			MP_WritePhyUshort(sc, 0x17, 0x7800);
			MP_WritePhyUshort(sc, 0x19, 0x3246);
		}

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x02)&0xF8FF;
		Data |= 0x0100;
		MP_WritePhyUshort(sc, 0x02, Data);
		Data = MP_ReadPhyUshort(sc, 0x03)&0x1FFF;
		MP_WritePhyUshort(sc, 0x03, Data);

		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);

		MP_WritePhyUshort(sc, 0x1f, 0x0005);
		MP_WritePhyUshort(sc, 0x05, 0x001B);
		if(MP_ReadPhyUshort(sc, 0x06)==0xBF00)
		{
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0xfff6);
			MP_WritePhyUshort(sc, 0x06, 0x0080);
			MP_WritePhyUshort(sc, 0x05, 0x8000);
			MP_WritePhyUshort(sc, 0x06, 0xf8f9);
			MP_WritePhyUshort(sc, 0x06, 0xfaef);
			MP_WritePhyUshort(sc, 0x06, 0x59ee);
			MP_WritePhyUshort(sc, 0x06, 0xf8ea);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0xf8eb);
			MP_WritePhyUshort(sc, 0x06, 0x00e0);
			MP_WritePhyUshort(sc, 0x06, 0xf87c);
			MP_WritePhyUshort(sc, 0x06, 0xe1f8);
			MP_WritePhyUshort(sc, 0x06, 0x7d59);
			MP_WritePhyUshort(sc, 0x06, 0x0fef);
			MP_WritePhyUshort(sc, 0x06, 0x0139);
			MP_WritePhyUshort(sc, 0x06, 0x029e);
			MP_WritePhyUshort(sc, 0x06, 0x06ef);
			MP_WritePhyUshort(sc, 0x06, 0x1039);
			MP_WritePhyUshort(sc, 0x06, 0x089f);
			MP_WritePhyUshort(sc, 0x06, 0x2aee);
			MP_WritePhyUshort(sc, 0x06, 0xf8ea);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0xf8eb);
			MP_WritePhyUshort(sc, 0x06, 0x01e0);
			MP_WritePhyUshort(sc, 0x06, 0xf87c);
			MP_WritePhyUshort(sc, 0x06, 0xe1f8);
			MP_WritePhyUshort(sc, 0x06, 0x7d58);
			MP_WritePhyUshort(sc, 0x06, 0x409e);
			MP_WritePhyUshort(sc, 0x06, 0x0f39);
			MP_WritePhyUshort(sc, 0x06, 0x46aa);
			MP_WritePhyUshort(sc, 0x06, 0x0bbf);
			MP_WritePhyUshort(sc, 0x06, 0x8290);
			MP_WritePhyUshort(sc, 0x06, 0xd682);
			MP_WritePhyUshort(sc, 0x06, 0x9802);
			MP_WritePhyUshort(sc, 0x06, 0x014f);
			MP_WritePhyUshort(sc, 0x06, 0xae09);
			MP_WritePhyUshort(sc, 0x06, 0xbf82);
			MP_WritePhyUshort(sc, 0x06, 0x98d6);
			MP_WritePhyUshort(sc, 0x06, 0x82a0);
			MP_WritePhyUshort(sc, 0x06, 0x0201);
			MP_WritePhyUshort(sc, 0x06, 0x4fef);
			MP_WritePhyUshort(sc, 0x06, 0x95fe);
			MP_WritePhyUshort(sc, 0x06, 0xfdfc);
			MP_WritePhyUshort(sc, 0x06, 0x05f8);
			MP_WritePhyUshort(sc, 0x06, 0xf9fa);
			MP_WritePhyUshort(sc, 0x06, 0xeef8);
			MP_WritePhyUshort(sc, 0x06, 0xea00);
			MP_WritePhyUshort(sc, 0x06, 0xeef8);
			MP_WritePhyUshort(sc, 0x06, 0xeb00);
			MP_WritePhyUshort(sc, 0x06, 0xe2f8);
			MP_WritePhyUshort(sc, 0x06, 0x7ce3);
			MP_WritePhyUshort(sc, 0x06, 0xf87d);
			MP_WritePhyUshort(sc, 0x06, 0xa511);
			MP_WritePhyUshort(sc, 0x06, 0x1112);
			MP_WritePhyUshort(sc, 0x06, 0xd240);
			MP_WritePhyUshort(sc, 0x06, 0xd644);
			MP_WritePhyUshort(sc, 0x06, 0x4402);
			MP_WritePhyUshort(sc, 0x06, 0x8217);
			MP_WritePhyUshort(sc, 0x06, 0xd2a0);
			MP_WritePhyUshort(sc, 0x06, 0xd6aa);
			MP_WritePhyUshort(sc, 0x06, 0xaa02);
			MP_WritePhyUshort(sc, 0x06, 0x8217);
			MP_WritePhyUshort(sc, 0x06, 0xae0f);
			MP_WritePhyUshort(sc, 0x06, 0xa544);
			MP_WritePhyUshort(sc, 0x06, 0x4402);
			MP_WritePhyUshort(sc, 0x06, 0xae4d);
			MP_WritePhyUshort(sc, 0x06, 0xa5aa);
			MP_WritePhyUshort(sc, 0x06, 0xaa02);
			MP_WritePhyUshort(sc, 0x06, 0xae47);
			MP_WritePhyUshort(sc, 0x06, 0xaf82);
			MP_WritePhyUshort(sc, 0x06, 0x13ee);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0x0fee);
			MP_WritePhyUshort(sc, 0x06, 0x834c);
			MP_WritePhyUshort(sc, 0x06, 0x0fee);
			MP_WritePhyUshort(sc, 0x06, 0x834f);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0x8351);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0x834a);
			MP_WritePhyUshort(sc, 0x06, 0xffee);
			MP_WritePhyUshort(sc, 0x06, 0x834b);
			MP_WritePhyUshort(sc, 0x06, 0xffe0);
			MP_WritePhyUshort(sc, 0x06, 0x8330);
			MP_WritePhyUshort(sc, 0x06, 0xe183);
			MP_WritePhyUshort(sc, 0x06, 0x3158);
			MP_WritePhyUshort(sc, 0x06, 0xfee4);
			MP_WritePhyUshort(sc, 0x06, 0xf88a);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x8be0);
			MP_WritePhyUshort(sc, 0x06, 0x8332);
			MP_WritePhyUshort(sc, 0x06, 0xe183);
			MP_WritePhyUshort(sc, 0x06, 0x3359);
			MP_WritePhyUshort(sc, 0x06, 0x0fe2);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0x0c24);
			MP_WritePhyUshort(sc, 0x06, 0x5af0);
			MP_WritePhyUshort(sc, 0x06, 0x1e12);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x8ce5);
			MP_WritePhyUshort(sc, 0x06, 0xf88d);
			MP_WritePhyUshort(sc, 0x06, 0xaf82);
			MP_WritePhyUshort(sc, 0x06, 0x13e0);
			MP_WritePhyUshort(sc, 0x06, 0x834f);
			MP_WritePhyUshort(sc, 0x06, 0x10e4);
			MP_WritePhyUshort(sc, 0x06, 0x834f);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x009f);
			MP_WritePhyUshort(sc, 0x06, 0x0ae0);
			MP_WritePhyUshort(sc, 0x06, 0x834f);
			MP_WritePhyUshort(sc, 0x06, 0xa010);
			MP_WritePhyUshort(sc, 0x06, 0xa5ee);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x01e0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7805);
			MP_WritePhyUshort(sc, 0x06, 0x9e9a);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x049e);
			MP_WritePhyUshort(sc, 0x06, 0x10e0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7803);
			MP_WritePhyUshort(sc, 0x06, 0x9e0f);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x019e);
			MP_WritePhyUshort(sc, 0x06, 0x05ae);
			MP_WritePhyUshort(sc, 0x06, 0x0caf);
			MP_WritePhyUshort(sc, 0x06, 0x81f8);
			MP_WritePhyUshort(sc, 0x06, 0xaf81);
			MP_WritePhyUshort(sc, 0x06, 0xa3af);
			MP_WritePhyUshort(sc, 0x06, 0x81dc);
			MP_WritePhyUshort(sc, 0x06, 0xaf82);
			MP_WritePhyUshort(sc, 0x06, 0x13ee);
			MP_WritePhyUshort(sc, 0x06, 0x8348);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0x8349);
			MP_WritePhyUshort(sc, 0x06, 0x00e0);
			MP_WritePhyUshort(sc, 0x06, 0x8351);
			MP_WritePhyUshort(sc, 0x06, 0x10e4);
			MP_WritePhyUshort(sc, 0x06, 0x8351);
			MP_WritePhyUshort(sc, 0x06, 0x5801);
			MP_WritePhyUshort(sc, 0x06, 0x9fea);
			MP_WritePhyUshort(sc, 0x06, 0xd000);
			MP_WritePhyUshort(sc, 0x06, 0xd180);
			MP_WritePhyUshort(sc, 0x06, 0x1f66);
			MP_WritePhyUshort(sc, 0x06, 0xe2f8);
			MP_WritePhyUshort(sc, 0x06, 0xeae3);
			MP_WritePhyUshort(sc, 0x06, 0xf8eb);
			MP_WritePhyUshort(sc, 0x06, 0x5af8);
			MP_WritePhyUshort(sc, 0x06, 0x1e20);
			MP_WritePhyUshort(sc, 0x06, 0xe6f8);
			MP_WritePhyUshort(sc, 0x06, 0xeae5);
			MP_WritePhyUshort(sc, 0x06, 0xf8eb);
			MP_WritePhyUshort(sc, 0x06, 0xd302);
			MP_WritePhyUshort(sc, 0x06, 0xb3fe);
			MP_WritePhyUshort(sc, 0x06, 0xe2f8);
			MP_WritePhyUshort(sc, 0x06, 0x7cef);
			MP_WritePhyUshort(sc, 0x06, 0x325b);
			MP_WritePhyUshort(sc, 0x06, 0x80e3);
			MP_WritePhyUshort(sc, 0x06, 0xf87d);
			MP_WritePhyUshort(sc, 0x06, 0x9e03);
			MP_WritePhyUshort(sc, 0x06, 0x7dff);
			MP_WritePhyUshort(sc, 0x06, 0xff0d);
			MP_WritePhyUshort(sc, 0x06, 0x581c);
			MP_WritePhyUshort(sc, 0x06, 0x551a);
			MP_WritePhyUshort(sc, 0x06, 0x6511);
			MP_WritePhyUshort(sc, 0x06, 0xa190);
			MP_WritePhyUshort(sc, 0x06, 0xd3e2);
			MP_WritePhyUshort(sc, 0x06, 0x8348);
			MP_WritePhyUshort(sc, 0x06, 0xe383);
			MP_WritePhyUshort(sc, 0x06, 0x491b);
			MP_WritePhyUshort(sc, 0x06, 0x56ab);
			MP_WritePhyUshort(sc, 0x06, 0x08ef);
			MP_WritePhyUshort(sc, 0x06, 0x56e6);
			MP_WritePhyUshort(sc, 0x06, 0x8348);
			MP_WritePhyUshort(sc, 0x06, 0xe783);
			MP_WritePhyUshort(sc, 0x06, 0x4910);
			MP_WritePhyUshort(sc, 0x06, 0xd180);
			MP_WritePhyUshort(sc, 0x06, 0x1f66);
			MP_WritePhyUshort(sc, 0x06, 0xa004);
			MP_WritePhyUshort(sc, 0x06, 0xb9e2);
			MP_WritePhyUshort(sc, 0x06, 0x8348);
			MP_WritePhyUshort(sc, 0x06, 0xe383);
			MP_WritePhyUshort(sc, 0x06, 0x49ef);
			MP_WritePhyUshort(sc, 0x06, 0x65e2);
			MP_WritePhyUshort(sc, 0x06, 0x834a);
			MP_WritePhyUshort(sc, 0x06, 0xe383);
			MP_WritePhyUshort(sc, 0x06, 0x4b1b);
			MP_WritePhyUshort(sc, 0x06, 0x56aa);
			MP_WritePhyUshort(sc, 0x06, 0x0eef);
			MP_WritePhyUshort(sc, 0x06, 0x56e6);
			MP_WritePhyUshort(sc, 0x06, 0x834a);
			MP_WritePhyUshort(sc, 0x06, 0xe783);
			MP_WritePhyUshort(sc, 0x06, 0x4be2);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0xe683);
			MP_WritePhyUshort(sc, 0x06, 0x4ce0);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0xa000);
			MP_WritePhyUshort(sc, 0x06, 0x0caf);
			MP_WritePhyUshort(sc, 0x06, 0x81dc);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4d10);
			MP_WritePhyUshort(sc, 0x06, 0xe483);
			MP_WritePhyUshort(sc, 0x06, 0x4dae);
			MP_WritePhyUshort(sc, 0x06, 0x0480);
			MP_WritePhyUshort(sc, 0x06, 0xe483);
			MP_WritePhyUshort(sc, 0x06, 0x4de0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7803);
			MP_WritePhyUshort(sc, 0x06, 0x9e0b);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x049e);
			MP_WritePhyUshort(sc, 0x06, 0x04ee);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x02e0);
			MP_WritePhyUshort(sc, 0x06, 0x8332);
			MP_WritePhyUshort(sc, 0x06, 0xe183);
			MP_WritePhyUshort(sc, 0x06, 0x3359);
			MP_WritePhyUshort(sc, 0x06, 0x0fe2);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0x0c24);
			MP_WritePhyUshort(sc, 0x06, 0x5af0);
			MP_WritePhyUshort(sc, 0x06, 0x1e12);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x8ce5);
			MP_WritePhyUshort(sc, 0x06, 0xf88d);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x30e1);
			MP_WritePhyUshort(sc, 0x06, 0x8331);
			MP_WritePhyUshort(sc, 0x06, 0x6801);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x8ae5);
			MP_WritePhyUshort(sc, 0x06, 0xf88b);
			MP_WritePhyUshort(sc, 0x06, 0xae37);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4e03);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4ce1);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0x1b01);
			MP_WritePhyUshort(sc, 0x06, 0x9e04);
			MP_WritePhyUshort(sc, 0x06, 0xaaa1);
			MP_WritePhyUshort(sc, 0x06, 0xaea8);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4e04);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4f00);
			MP_WritePhyUshort(sc, 0x06, 0xaeab);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4f78);
			MP_WritePhyUshort(sc, 0x06, 0x039f);
			MP_WritePhyUshort(sc, 0x06, 0x14ee);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x05d2);
			MP_WritePhyUshort(sc, 0x06, 0x40d6);
			MP_WritePhyUshort(sc, 0x06, 0x5554);
			MP_WritePhyUshort(sc, 0x06, 0x0282);
			MP_WritePhyUshort(sc, 0x06, 0x17d2);
			MP_WritePhyUshort(sc, 0x06, 0xa0d6);
			MP_WritePhyUshort(sc, 0x06, 0xba00);
			MP_WritePhyUshort(sc, 0x06, 0x0282);
			MP_WritePhyUshort(sc, 0x06, 0x17fe);
			MP_WritePhyUshort(sc, 0x06, 0xfdfc);
			MP_WritePhyUshort(sc, 0x06, 0x05f8);
			MP_WritePhyUshort(sc, 0x06, 0xe0f8);
			MP_WritePhyUshort(sc, 0x06, 0x60e1);
			MP_WritePhyUshort(sc, 0x06, 0xf861);
			MP_WritePhyUshort(sc, 0x06, 0x6802);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x60e5);
			MP_WritePhyUshort(sc, 0x06, 0xf861);
			MP_WritePhyUshort(sc, 0x06, 0xe0f8);
			MP_WritePhyUshort(sc, 0x06, 0x48e1);
			MP_WritePhyUshort(sc, 0x06, 0xf849);
			MP_WritePhyUshort(sc, 0x06, 0x580f);
			MP_WritePhyUshort(sc, 0x06, 0x1e02);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x48e5);
			MP_WritePhyUshort(sc, 0x06, 0xf849);
			MP_WritePhyUshort(sc, 0x06, 0xd000);
			MP_WritePhyUshort(sc, 0x06, 0x0282);
			MP_WritePhyUshort(sc, 0x06, 0x5bbf);
			MP_WritePhyUshort(sc, 0x06, 0x8350);
			MP_WritePhyUshort(sc, 0x06, 0xef46);
			MP_WritePhyUshort(sc, 0x06, 0xdc19);
			MP_WritePhyUshort(sc, 0x06, 0xddd0);
			MP_WritePhyUshort(sc, 0x06, 0x0102);
			MP_WritePhyUshort(sc, 0x06, 0x825b);
			MP_WritePhyUshort(sc, 0x06, 0x0282);
			MP_WritePhyUshort(sc, 0x06, 0x77e0);
			MP_WritePhyUshort(sc, 0x06, 0xf860);
			MP_WritePhyUshort(sc, 0x06, 0xe1f8);
			MP_WritePhyUshort(sc, 0x06, 0x6158);
			MP_WritePhyUshort(sc, 0x06, 0xfde4);
			MP_WritePhyUshort(sc, 0x06, 0xf860);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x61fc);
			MP_WritePhyUshort(sc, 0x06, 0x04f9);
			MP_WritePhyUshort(sc, 0x06, 0xfafb);
			MP_WritePhyUshort(sc, 0x06, 0xc6bf);
			MP_WritePhyUshort(sc, 0x06, 0xf840);
			MP_WritePhyUshort(sc, 0x06, 0xbe83);
			MP_WritePhyUshort(sc, 0x06, 0x50a0);
			MP_WritePhyUshort(sc, 0x06, 0x0101);
			MP_WritePhyUshort(sc, 0x06, 0x071b);
			MP_WritePhyUshort(sc, 0x06, 0x89cf);
			MP_WritePhyUshort(sc, 0x06, 0xd208);
			MP_WritePhyUshort(sc, 0x06, 0xebdb);
			MP_WritePhyUshort(sc, 0x06, 0x19b2);
			MP_WritePhyUshort(sc, 0x06, 0xfbff);
			MP_WritePhyUshort(sc, 0x06, 0xfefd);
			MP_WritePhyUshort(sc, 0x06, 0x04f8);
			MP_WritePhyUshort(sc, 0x06, 0xe0f8);
			MP_WritePhyUshort(sc, 0x06, 0x48e1);
			MP_WritePhyUshort(sc, 0x06, 0xf849);
			MP_WritePhyUshort(sc, 0x06, 0x6808);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x48e5);
			MP_WritePhyUshort(sc, 0x06, 0xf849);
			MP_WritePhyUshort(sc, 0x06, 0x58f7);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x48e5);
			MP_WritePhyUshort(sc, 0x06, 0xf849);
			MP_WritePhyUshort(sc, 0x06, 0xfc04);
			MP_WritePhyUshort(sc, 0x06, 0x4d20);
			MP_WritePhyUshort(sc, 0x06, 0x0002);
			MP_WritePhyUshort(sc, 0x06, 0x4e22);
			MP_WritePhyUshort(sc, 0x06, 0x0002);
			MP_WritePhyUshort(sc, 0x06, 0x4ddf);
			MP_WritePhyUshort(sc, 0x06, 0xff01);
			MP_WritePhyUshort(sc, 0x06, 0x4edd);
			MP_WritePhyUshort(sc, 0x06, 0xff01);
			MP_WritePhyUshort(sc, 0x06, 0xf8fa);
			MP_WritePhyUshort(sc, 0x06, 0xfbef);
			MP_WritePhyUshort(sc, 0x06, 0x79bf);
			MP_WritePhyUshort(sc, 0x06, 0xf822);
			MP_WritePhyUshort(sc, 0x06, 0xd819);
			MP_WritePhyUshort(sc, 0x06, 0xd958);
			MP_WritePhyUshort(sc, 0x06, 0x849f);
			MP_WritePhyUshort(sc, 0x06, 0x09bf);
			MP_WritePhyUshort(sc, 0x06, 0x82be);
			MP_WritePhyUshort(sc, 0x06, 0xd682);
			MP_WritePhyUshort(sc, 0x06, 0xc602);
			MP_WritePhyUshort(sc, 0x06, 0x014f);
			MP_WritePhyUshort(sc, 0x06, 0xef97);
			MP_WritePhyUshort(sc, 0x06, 0xfffe);
			MP_WritePhyUshort(sc, 0x06, 0xfc05);
			MP_WritePhyUshort(sc, 0x06, 0x17ff);
			MP_WritePhyUshort(sc, 0x06, 0xfe01);
			MP_WritePhyUshort(sc, 0x06, 0x1700);
			MP_WritePhyUshort(sc, 0x06, 0x0102);
			MP_WritePhyUshort(sc, 0x05, 0x83d8);
			MP_WritePhyUshort(sc, 0x06, 0x8051);
			MP_WritePhyUshort(sc, 0x05, 0x83d6);
			MP_WritePhyUshort(sc, 0x06, 0x82a0);
			MP_WritePhyUshort(sc, 0x05, 0x83d4);
			MP_WritePhyUshort(sc, 0x06, 0x8000);
			MP_WritePhyUshort(sc, 0x02, 0x2010);
			MP_WritePhyUshort(sc, 0x03, 0xdc00);
			MP_WritePhyUshort(sc, 0x1f, 0x0000);
			MP_WritePhyUshort(sc, 0x0b, 0x0600);
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0xfff6);
			MP_WritePhyUshort(sc, 0x06, 0x00fc);
			MP_WritePhyUshort(sc, 0x1f, 0x0000);
		}

	} else if (sc->rl_type == MACFG_32) {
		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x06, 0x4064);
		MP_WritePhyUshort(sc, 0x07, 0x2863);
		MP_WritePhyUshort(sc, 0x08, 0x059C);
		MP_WritePhyUshort(sc, 0x09, 0x26B4);
		MP_WritePhyUshort(sc, 0x0A, 0x6A19);
		MP_WritePhyUshort(sc, 0x0B, 0xBCC0);
		MP_WritePhyUshort(sc, 0x10, 0xF06D);
		MP_WritePhyUshort(sc, 0x14, 0x7F68);
		MP_WritePhyUshort(sc, 0x18, 0x7FD9);
		MP_WritePhyUshort(sc, 0x1C, 0xF0FF);
		MP_WritePhyUshort(sc, 0x1D, 0x3D9C);
		MP_WritePhyUshort(sc, 0x1F, 0x0003);
		MP_WritePhyUshort(sc, 0x12, 0xF49F);
		MP_WritePhyUshort(sc, 0x13, 0x070B);
		MP_WritePhyUshort(sc, 0x1A, 0x05AD);
		MP_WritePhyUshort(sc, 0x14, 0x94C0);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x06, 0x5571);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x05, 0x2642);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x02, 0xC107);
		MP_WritePhyUshort(sc, 0x03, 0x1002);

		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x16, 0x0CC0);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x0F, 0x0017);

		MP_WritePhyUshort(sc, 0x1F, 0x0005);
		MP_WritePhyUshort(sc, 0x05, 0x8200);
		MP_WritePhyUshort(sc, 0x06, 0xF8F9);
		MP_WritePhyUshort(sc, 0x06, 0xFAEF);
		MP_WritePhyUshort(sc, 0x06, 0x59EE);
		MP_WritePhyUshort(sc, 0x06, 0xF8EA);
		MP_WritePhyUshort(sc, 0x06, 0x00EE);
		MP_WritePhyUshort(sc, 0x06, 0xF8EB);
		MP_WritePhyUshort(sc, 0x06, 0x00E0);
		MP_WritePhyUshort(sc, 0x06, 0xF87C);
		MP_WritePhyUshort(sc, 0x06, 0xE1F8);
		MP_WritePhyUshort(sc, 0x06, 0x7D59);
		MP_WritePhyUshort(sc, 0x06, 0x0FEF);
		MP_WritePhyUshort(sc, 0x06, 0x0139);
		MP_WritePhyUshort(sc, 0x06, 0x029E);
		MP_WritePhyUshort(sc, 0x06, 0x06EF);
		MP_WritePhyUshort(sc, 0x06, 0x1039);
		MP_WritePhyUshort(sc, 0x06, 0x089F);
		MP_WritePhyUshort(sc, 0x06, 0x2AEE);
		MP_WritePhyUshort(sc, 0x06, 0xF8EA);
		MP_WritePhyUshort(sc, 0x06, 0x00EE);
		MP_WritePhyUshort(sc, 0x06, 0xF8EB);
		MP_WritePhyUshort(sc, 0x06, 0x01E0);
		MP_WritePhyUshort(sc, 0x06, 0xF87C);
		MP_WritePhyUshort(sc, 0x06, 0xE1F8);
		MP_WritePhyUshort(sc, 0x06, 0x7D58);
		MP_WritePhyUshort(sc, 0x06, 0x409E);
		MP_WritePhyUshort(sc, 0x06, 0x0F39);
		MP_WritePhyUshort(sc, 0x06, 0x46AA);
		MP_WritePhyUshort(sc, 0x06, 0x0BBF);
		MP_WritePhyUshort(sc, 0x06, 0x8251);
		MP_WritePhyUshort(sc, 0x06, 0xD682);
		MP_WritePhyUshort(sc, 0x06, 0x5902);
		MP_WritePhyUshort(sc, 0x06, 0x014F);
		MP_WritePhyUshort(sc, 0x06, 0xAE09);
		MP_WritePhyUshort(sc, 0x06, 0xBF82);
		MP_WritePhyUshort(sc, 0x06, 0x59D6);
		MP_WritePhyUshort(sc, 0x06, 0x8261);
		MP_WritePhyUshort(sc, 0x06, 0x0201);
		MP_WritePhyUshort(sc, 0x06, 0x4FEF);
		MP_WritePhyUshort(sc, 0x06, 0x95FE);
		MP_WritePhyUshort(sc, 0x06, 0xFDFC);
		MP_WritePhyUshort(sc, 0x06, 0x054D);
		MP_WritePhyUshort(sc, 0x06, 0x2000);
		MP_WritePhyUshort(sc, 0x06, 0x024E);
		MP_WritePhyUshort(sc, 0x06, 0x2200);
		MP_WritePhyUshort(sc, 0x06, 0x024D);
		MP_WritePhyUshort(sc, 0x06, 0xDFFF);
		MP_WritePhyUshort(sc, 0x06, 0x014E);
		MP_WritePhyUshort(sc, 0x06, 0xDDFF);
		MP_WritePhyUshort(sc, 0x06, 0x0100);
		MP_WritePhyUshort(sc, 0x02, 0x6010);
		MP_WritePhyUshort(sc, 0x05, 0xFFF6);
		MP_WritePhyUshort(sc, 0x06, 0x00EC);
		MP_WritePhyUshort(sc, 0x05, 0x83D4);
		MP_WritePhyUshort(sc, 0x06, 0x8200);

	} else if (sc->rl_type == MACFG_33) {
		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x06, 0x4064);
		MP_WritePhyUshort(sc, 0x07, 0x2863);
		MP_WritePhyUshort(sc, 0x08, 0x059C);
		MP_WritePhyUshort(sc, 0x09, 0x26B4);
		MP_WritePhyUshort(sc, 0x0A, 0x6A19);
		MP_WritePhyUshort(sc, 0x0B, 0xBCC0);
		MP_WritePhyUshort(sc, 0x10, 0xF06D);
		MP_WritePhyUshort(sc, 0x14, 0x7F68);
		MP_WritePhyUshort(sc, 0x18, 0x7FD9);
		MP_WritePhyUshort(sc, 0x1C, 0xF0FF);
		MP_WritePhyUshort(sc, 0x1D, 0x3D9C);
		MP_WritePhyUshort(sc, 0x1F, 0x0003);
		MP_WritePhyUshort(sc, 0x12, 0xF49F);
		MP_WritePhyUshort(sc, 0x13, 0x070B);
		MP_WritePhyUshort(sc, 0x1A, 0x05AD);
		MP_WritePhyUshort(sc, 0x14, 0x94C0);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		MP_WritePhyUshort(sc, 0x06, 0x5561);
		MP_WritePhyUshort(sc, 0x1f, 0x0005);
		MP_WritePhyUshort(sc, 0x05, 0x8332);
		MP_WritePhyUshort(sc, 0x06, 0x5561);

		if(MP_ReadEfuse(sc, 0x01)==0xB1)
		{
			MP_WritePhyUshort(sc, 0x1f, 0x0002);
			MP_WritePhyUshort(sc, 0x05, 0x669a);
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0x8330);
			MP_WritePhyUshort(sc, 0x06, 0x669a);
			MP_WritePhyUshort(sc, 0x1f, 0x0002);
			Data = MP_ReadPhyUshort(sc, 0x0D);
			if((Data & 0x00FF)!=0x006C)
			{
				Data &= 0xFF00;
				MP_WritePhyUshort(sc, 0x0D, 0x0065|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0066|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0067|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0068|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x0069|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x006A|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x006B|Data);
				MP_WritePhyUshort(sc, 0x0D, 0x006C|Data);
			}
		}
		else
		{
			MP_WritePhyUshort(sc, 0x1f, 0x0002);
			MP_WritePhyUshort(sc, 0x05, 0x2642);
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0x8330);
			MP_WritePhyUshort(sc, 0x06, 0x2642);
		}

		MP_WritePhyUshort(sc, 0x1F, 0x0005);
		Data = MP_ReadPhyUshort(sc, 0x01);
		Data &= ~0x0200;
		MP_WritePhyUshort(sc, 0x01, Data);
		MP_WritePhyUshort(sc, 0x1F, 0x0000);
		MP_WritePhyUshort(sc, 0x16, 0x5101);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x05);
		Data &= 0xF0FF;
		if(CSR_READ_1(sc, RL_PHY_STATUS)&0x10)
		{
			switch(Data)
			{
				case 0x609A:
					MP_WritePhyUshort(sc, 0x15, 0x3928);
					MP_WritePhyUshort(sc, 0x16, 0x794c);
					MP_WritePhyUshort(sc, 0x17, 0x79b4);
					break;
				case 0x6062:
					MP_WritePhyUshort(sc, 0x15, 0x3628);
					MP_WritePhyUshort(sc, 0x16, 0x764c);
					MP_WritePhyUshort(sc, 0x17, 0x76b4);
					break;
				case 0x2042:
					MP_WritePhyUshort(sc, 0x15, 0x3428);
					MP_WritePhyUshort(sc, 0x16, 0x344c);
					MP_WritePhyUshort(sc, 0x17, 0x34b4);
					break;
			}
			MP_WritePhyUshort(sc, 0x1F, 0x0003);
			MP_WritePhyUshort(sc, 0x17, 0x7bc0);
			MP_WritePhyUshort(sc, 0x18, 0xff46);
			MP_WritePhyUshort(sc, 0x19, 0x3229);
		}
		else
		{
			switch(Data)
			{
				case 0x609A:
					MP_WritePhyUshort(sc, 0x15, 0x3900);
					MP_WritePhyUshort(sc, 0x16, 0x7924);
					MP_WritePhyUshort(sc, 0x17, 0x7948);
					break;
				case 0x6062:
					MP_WritePhyUshort(sc, 0x15, 0x3600);
					MP_WritePhyUshort(sc, 0x16, 0x7624);
					MP_WritePhyUshort(sc, 0x17, 0x7648);
					break;
				case 0x2042:
					MP_WritePhyUshort(sc, 0x15, 0x3400);
					MP_WritePhyUshort(sc, 0x16, 0x3424);
					MP_WritePhyUshort(sc, 0x17, 0x3448);
					break;
			}
			MP_WritePhyUshort(sc, 0x1f, 0x0003);
			MP_WritePhyUshort(sc, 0x10, 0xAC7C);
			MP_WritePhyUshort(sc, 0x17, 0x7800);
			MP_WritePhyUshort(sc, 0x19, 0x3246);
		}

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x02)&0xF8FF;
		Data |= 0x0100;
		MP_WritePhyUshort(sc, 0x02, Data);
		Data = MP_ReadPhyUshort(sc, 0x03)&0x1FFF;
		MP_WritePhyUshort(sc, 0x03, Data);

		MP_WritePhyUshort(sc, 0x1F, 0x0001);
		MP_WritePhyUshort(sc, 0x17, 0x0CC0);

		MP_WritePhyUshort(sc, 0x1F, 0x0002);
		Data = MP_ReadPhyUshort(sc, 0x0F);
		Data |= 0x0017;
		MP_WritePhyUshort(sc, 0x0F, Data);

		MP_WritePhyUshort(sc, 0x1f, 0x0005);
		MP_WritePhyUshort(sc, 0x05, 0x001B);
		if(MP_ReadPhyUshort(sc, 0x06)==0xB300)
		{
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0xfff6);
			MP_WritePhyUshort(sc, 0x06, 0x0080);
			MP_WritePhyUshort(sc, 0x05, 0x8000);
			MP_WritePhyUshort(sc, 0x06, 0xf8f9);
			MP_WritePhyUshort(sc, 0x06, 0xfaee);
			MP_WritePhyUshort(sc, 0x06, 0xf8ea);
			MP_WritePhyUshort(sc, 0x06, 0x00ee);
			MP_WritePhyUshort(sc, 0x06, 0xf8eb);
			MP_WritePhyUshort(sc, 0x06, 0x00e2);
			MP_WritePhyUshort(sc, 0x06, 0xf87c);
			MP_WritePhyUshort(sc, 0x06, 0xe3f8);
			MP_WritePhyUshort(sc, 0x06, 0x7da5);
			MP_WritePhyUshort(sc, 0x06, 0x1111);
			MP_WritePhyUshort(sc, 0x06, 0x12d2);
			MP_WritePhyUshort(sc, 0x06, 0x40d6);
			MP_WritePhyUshort(sc, 0x06, 0x4444);
			MP_WritePhyUshort(sc, 0x06, 0x0281);
			MP_WritePhyUshort(sc, 0x06, 0xc6d2);
			MP_WritePhyUshort(sc, 0x06, 0xa0d6);
			MP_WritePhyUshort(sc, 0x06, 0xaaaa);
			MP_WritePhyUshort(sc, 0x06, 0x0281);
			MP_WritePhyUshort(sc, 0x06, 0xc6ae);
			MP_WritePhyUshort(sc, 0x06, 0x0fa5);
			MP_WritePhyUshort(sc, 0x06, 0x4444);
			MP_WritePhyUshort(sc, 0x06, 0x02ae);
			MP_WritePhyUshort(sc, 0x06, 0x4da5);
			MP_WritePhyUshort(sc, 0x06, 0xaaaa);
			MP_WritePhyUshort(sc, 0x06, 0x02ae);
			MP_WritePhyUshort(sc, 0x06, 0x47af);
			MP_WritePhyUshort(sc, 0x06, 0x81c2);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4e00);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4d0f);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4c0f);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4f00);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x5100);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4aff);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4bff);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x30e1);
			MP_WritePhyUshort(sc, 0x06, 0x8331);
			MP_WritePhyUshort(sc, 0x06, 0x58fe);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x8ae5);
			MP_WritePhyUshort(sc, 0x06, 0xf88b);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x32e1);
			MP_WritePhyUshort(sc, 0x06, 0x8333);
			MP_WritePhyUshort(sc, 0x06, 0x590f);
			MP_WritePhyUshort(sc, 0x06, 0xe283);
			MP_WritePhyUshort(sc, 0x06, 0x4d0c);
			MP_WritePhyUshort(sc, 0x06, 0x245a);
			MP_WritePhyUshort(sc, 0x06, 0xf01e);
			MP_WritePhyUshort(sc, 0x06, 0x12e4);
			MP_WritePhyUshort(sc, 0x06, 0xf88c);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x8daf);
			MP_WritePhyUshort(sc, 0x06, 0x81c2);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4f10);
			MP_WritePhyUshort(sc, 0x06, 0xe483);
			MP_WritePhyUshort(sc, 0x06, 0x4fe0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7800);
			MP_WritePhyUshort(sc, 0x06, 0x9f0a);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4fa0);
			MP_WritePhyUshort(sc, 0x06, 0x10a5);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4e01);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x059e);
			MP_WritePhyUshort(sc, 0x06, 0x9ae0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7804);
			MP_WritePhyUshort(sc, 0x06, 0x9e10);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x039e);
			MP_WritePhyUshort(sc, 0x06, 0x0fe0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7801);
			MP_WritePhyUshort(sc, 0x06, 0x9e05);
			MP_WritePhyUshort(sc, 0x06, 0xae0c);
			MP_WritePhyUshort(sc, 0x06, 0xaf81);
			MP_WritePhyUshort(sc, 0x06, 0xa7af);
			MP_WritePhyUshort(sc, 0x06, 0x8152);
			MP_WritePhyUshort(sc, 0x06, 0xaf81);
			MP_WritePhyUshort(sc, 0x06, 0x8baf);
			MP_WritePhyUshort(sc, 0x06, 0x81c2);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4800);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4900);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x5110);
			MP_WritePhyUshort(sc, 0x06, 0xe483);
			MP_WritePhyUshort(sc, 0x06, 0x5158);
			MP_WritePhyUshort(sc, 0x06, 0x019f);
			MP_WritePhyUshort(sc, 0x06, 0xead0);
			MP_WritePhyUshort(sc, 0x06, 0x00d1);
			MP_WritePhyUshort(sc, 0x06, 0x801f);
			MP_WritePhyUshort(sc, 0x06, 0x66e2);
			MP_WritePhyUshort(sc, 0x06, 0xf8ea);
			MP_WritePhyUshort(sc, 0x06, 0xe3f8);
			MP_WritePhyUshort(sc, 0x06, 0xeb5a);
			MP_WritePhyUshort(sc, 0x06, 0xf81e);
			MP_WritePhyUshort(sc, 0x06, 0x20e6);
			MP_WritePhyUshort(sc, 0x06, 0xf8ea);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0xebd3);
			MP_WritePhyUshort(sc, 0x06, 0x02b3);
			MP_WritePhyUshort(sc, 0x06, 0xfee2);
			MP_WritePhyUshort(sc, 0x06, 0xf87c);
			MP_WritePhyUshort(sc, 0x06, 0xef32);
			MP_WritePhyUshort(sc, 0x06, 0x5b80);
			MP_WritePhyUshort(sc, 0x06, 0xe3f8);
			MP_WritePhyUshort(sc, 0x06, 0x7d9e);
			MP_WritePhyUshort(sc, 0x06, 0x037d);
			MP_WritePhyUshort(sc, 0x06, 0xffff);
			MP_WritePhyUshort(sc, 0x06, 0x0d58);
			MP_WritePhyUshort(sc, 0x06, 0x1c55);
			MP_WritePhyUshort(sc, 0x06, 0x1a65);
			MP_WritePhyUshort(sc, 0x06, 0x11a1);
			MP_WritePhyUshort(sc, 0x06, 0x90d3);
			MP_WritePhyUshort(sc, 0x06, 0xe283);
			MP_WritePhyUshort(sc, 0x06, 0x48e3);
			MP_WritePhyUshort(sc, 0x06, 0x8349);
			MP_WritePhyUshort(sc, 0x06, 0x1b56);
			MP_WritePhyUshort(sc, 0x06, 0xab08);
			MP_WritePhyUshort(sc, 0x06, 0xef56);
			MP_WritePhyUshort(sc, 0x06, 0xe683);
			MP_WritePhyUshort(sc, 0x06, 0x48e7);
			MP_WritePhyUshort(sc, 0x06, 0x8349);
			MP_WritePhyUshort(sc, 0x06, 0x10d1);
			MP_WritePhyUshort(sc, 0x06, 0x801f);
			MP_WritePhyUshort(sc, 0x06, 0x66a0);
			MP_WritePhyUshort(sc, 0x06, 0x04b9);
			MP_WritePhyUshort(sc, 0x06, 0xe283);
			MP_WritePhyUshort(sc, 0x06, 0x48e3);
			MP_WritePhyUshort(sc, 0x06, 0x8349);
			MP_WritePhyUshort(sc, 0x06, 0xef65);
			MP_WritePhyUshort(sc, 0x06, 0xe283);
			MP_WritePhyUshort(sc, 0x06, 0x4ae3);
			MP_WritePhyUshort(sc, 0x06, 0x834b);
			MP_WritePhyUshort(sc, 0x06, 0x1b56);
			MP_WritePhyUshort(sc, 0x06, 0xaa0e);
			MP_WritePhyUshort(sc, 0x06, 0xef56);
			MP_WritePhyUshort(sc, 0x06, 0xe683);
			MP_WritePhyUshort(sc, 0x06, 0x4ae7);
			MP_WritePhyUshort(sc, 0x06, 0x834b);
			MP_WritePhyUshort(sc, 0x06, 0xe283);
			MP_WritePhyUshort(sc, 0x06, 0x4de6);
			MP_WritePhyUshort(sc, 0x06, 0x834c);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4da0);
			MP_WritePhyUshort(sc, 0x06, 0x000c);
			MP_WritePhyUshort(sc, 0x06, 0xaf81);
			MP_WritePhyUshort(sc, 0x06, 0x8be0);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0x10e4);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0xae04);
			MP_WritePhyUshort(sc, 0x06, 0x80e4);
			MP_WritePhyUshort(sc, 0x06, 0x834d);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x4e78);
			MP_WritePhyUshort(sc, 0x06, 0x039e);
			MP_WritePhyUshort(sc, 0x06, 0x0be0);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x7804);
			MP_WritePhyUshort(sc, 0x06, 0x9e04);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4e02);
			MP_WritePhyUshort(sc, 0x06, 0xe083);
			MP_WritePhyUshort(sc, 0x06, 0x32e1);
			MP_WritePhyUshort(sc, 0x06, 0x8333);
			MP_WritePhyUshort(sc, 0x06, 0x590f);
			MP_WritePhyUshort(sc, 0x06, 0xe283);
			MP_WritePhyUshort(sc, 0x06, 0x4d0c);
			MP_WritePhyUshort(sc, 0x06, 0x245a);
			MP_WritePhyUshort(sc, 0x06, 0xf01e);
			MP_WritePhyUshort(sc, 0x06, 0x12e4);
			MP_WritePhyUshort(sc, 0x06, 0xf88c);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x8de0);
			MP_WritePhyUshort(sc, 0x06, 0x8330);
			MP_WritePhyUshort(sc, 0x06, 0xe183);
			MP_WritePhyUshort(sc, 0x06, 0x3168);
			MP_WritePhyUshort(sc, 0x06, 0x01e4);
			MP_WritePhyUshort(sc, 0x06, 0xf88a);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x8bae);
			MP_WritePhyUshort(sc, 0x06, 0x37ee);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x03e0);
			MP_WritePhyUshort(sc, 0x06, 0x834c);
			MP_WritePhyUshort(sc, 0x06, 0xe183);
			MP_WritePhyUshort(sc, 0x06, 0x4d1b);
			MP_WritePhyUshort(sc, 0x06, 0x019e);
			MP_WritePhyUshort(sc, 0x06, 0x04aa);
			MP_WritePhyUshort(sc, 0x06, 0xa1ae);
			MP_WritePhyUshort(sc, 0x06, 0xa8ee);
			MP_WritePhyUshort(sc, 0x06, 0x834e);
			MP_WritePhyUshort(sc, 0x06, 0x04ee);
			MP_WritePhyUshort(sc, 0x06, 0x834f);
			MP_WritePhyUshort(sc, 0x06, 0x00ae);
			MP_WritePhyUshort(sc, 0x06, 0xabe0);
			MP_WritePhyUshort(sc, 0x06, 0x834f);
			MP_WritePhyUshort(sc, 0x06, 0x7803);
			MP_WritePhyUshort(sc, 0x06, 0x9f14);
			MP_WritePhyUshort(sc, 0x06, 0xee83);
			MP_WritePhyUshort(sc, 0x06, 0x4e05);
			MP_WritePhyUshort(sc, 0x06, 0xd240);
			MP_WritePhyUshort(sc, 0x06, 0xd655);
			MP_WritePhyUshort(sc, 0x06, 0x5402);
			MP_WritePhyUshort(sc, 0x06, 0x81c6);
			MP_WritePhyUshort(sc, 0x06, 0xd2a0);
			MP_WritePhyUshort(sc, 0x06, 0xd6ba);
			MP_WritePhyUshort(sc, 0x06, 0x0002);
			MP_WritePhyUshort(sc, 0x06, 0x81c6);
			MP_WritePhyUshort(sc, 0x06, 0xfefd);
			MP_WritePhyUshort(sc, 0x06, 0xfc05);
			MP_WritePhyUshort(sc, 0x06, 0xf8e0);
			MP_WritePhyUshort(sc, 0x06, 0xf860);
			MP_WritePhyUshort(sc, 0x06, 0xe1f8);
			MP_WritePhyUshort(sc, 0x06, 0x6168);
			MP_WritePhyUshort(sc, 0x06, 0x02e4);
			MP_WritePhyUshort(sc, 0x06, 0xf860);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x61e0);
			MP_WritePhyUshort(sc, 0x06, 0xf848);
			MP_WritePhyUshort(sc, 0x06, 0xe1f8);
			MP_WritePhyUshort(sc, 0x06, 0x4958);
			MP_WritePhyUshort(sc, 0x06, 0x0f1e);
			MP_WritePhyUshort(sc, 0x06, 0x02e4);
			MP_WritePhyUshort(sc, 0x06, 0xf848);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x49d0);
			MP_WritePhyUshort(sc, 0x06, 0x0002);
			MP_WritePhyUshort(sc, 0x06, 0x820a);
			MP_WritePhyUshort(sc, 0x06, 0xbf83);
			MP_WritePhyUshort(sc, 0x06, 0x50ef);
			MP_WritePhyUshort(sc, 0x06, 0x46dc);
			MP_WritePhyUshort(sc, 0x06, 0x19dd);
			MP_WritePhyUshort(sc, 0x06, 0xd001);
			MP_WritePhyUshort(sc, 0x06, 0x0282);
			MP_WritePhyUshort(sc, 0x06, 0x0a02);
			MP_WritePhyUshort(sc, 0x06, 0x8226);
			MP_WritePhyUshort(sc, 0x06, 0xe0f8);
			MP_WritePhyUshort(sc, 0x06, 0x60e1);
			MP_WritePhyUshort(sc, 0x06, 0xf861);
			MP_WritePhyUshort(sc, 0x06, 0x58fd);
			MP_WritePhyUshort(sc, 0x06, 0xe4f8);
			MP_WritePhyUshort(sc, 0x06, 0x60e5);
			MP_WritePhyUshort(sc, 0x06, 0xf861);
			MP_WritePhyUshort(sc, 0x06, 0xfc04);
			MP_WritePhyUshort(sc, 0x06, 0xf9fa);
			MP_WritePhyUshort(sc, 0x06, 0xfbc6);
			MP_WritePhyUshort(sc, 0x06, 0xbff8);
			MP_WritePhyUshort(sc, 0x06, 0x40be);
			MP_WritePhyUshort(sc, 0x06, 0x8350);
			MP_WritePhyUshort(sc, 0x06, 0xa001);
			MP_WritePhyUshort(sc, 0x06, 0x0107);
			MP_WritePhyUshort(sc, 0x06, 0x1b89);
			MP_WritePhyUshort(sc, 0x06, 0xcfd2);
			MP_WritePhyUshort(sc, 0x06, 0x08eb);
			MP_WritePhyUshort(sc, 0x06, 0xdb19);
			MP_WritePhyUshort(sc, 0x06, 0xb2fb);
			MP_WritePhyUshort(sc, 0x06, 0xfffe);
			MP_WritePhyUshort(sc, 0x06, 0xfd04);
			MP_WritePhyUshort(sc, 0x06, 0xf8e0);
			MP_WritePhyUshort(sc, 0x06, 0xf848);
			MP_WritePhyUshort(sc, 0x06, 0xe1f8);
			MP_WritePhyUshort(sc, 0x06, 0x4968);
			MP_WritePhyUshort(sc, 0x06, 0x08e4);
			MP_WritePhyUshort(sc, 0x06, 0xf848);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x4958);
			MP_WritePhyUshort(sc, 0x06, 0xf7e4);
			MP_WritePhyUshort(sc, 0x06, 0xf848);
			MP_WritePhyUshort(sc, 0x06, 0xe5f8);
			MP_WritePhyUshort(sc, 0x06, 0x49fc);
			MP_WritePhyUshort(sc, 0x06, 0x044d);
			MP_WritePhyUshort(sc, 0x06, 0x2000);
			MP_WritePhyUshort(sc, 0x06, 0x024e);
			MP_WritePhyUshort(sc, 0x06, 0x2200);
			MP_WritePhyUshort(sc, 0x06, 0x024d);
			MP_WritePhyUshort(sc, 0x06, 0xdfff);
			MP_WritePhyUshort(sc, 0x06, 0x014e);
			MP_WritePhyUshort(sc, 0x06, 0xddff);
			MP_WritePhyUshort(sc, 0x06, 0x01f8);
			MP_WritePhyUshort(sc, 0x06, 0xfafb);
			MP_WritePhyUshort(sc, 0x06, 0xef79);
			MP_WritePhyUshort(sc, 0x06, 0xbff8);
			MP_WritePhyUshort(sc, 0x06, 0x22d8);
			MP_WritePhyUshort(sc, 0x06, 0x19d9);
			MP_WritePhyUshort(sc, 0x06, 0x5884);
			MP_WritePhyUshort(sc, 0x06, 0x9f09);
			MP_WritePhyUshort(sc, 0x06, 0xbf82);
			MP_WritePhyUshort(sc, 0x06, 0x6dd6);
			MP_WritePhyUshort(sc, 0x06, 0x8275);
			MP_WritePhyUshort(sc, 0x06, 0x0201);
			MP_WritePhyUshort(sc, 0x06, 0x4fef);
			MP_WritePhyUshort(sc, 0x06, 0x97ff);
			MP_WritePhyUshort(sc, 0x06, 0xfefc);
			MP_WritePhyUshort(sc, 0x06, 0x0517);
			MP_WritePhyUshort(sc, 0x06, 0xfffe);
			MP_WritePhyUshort(sc, 0x06, 0x0117);
			MP_WritePhyUshort(sc, 0x06, 0x0001);
			MP_WritePhyUshort(sc, 0x06, 0x0200);
			MP_WritePhyUshort(sc, 0x05, 0x83d8);
			MP_WritePhyUshort(sc, 0x06, 0x8000);
			MP_WritePhyUshort(sc, 0x05, 0x83d6);
			MP_WritePhyUshort(sc, 0x06, 0x824f);
			MP_WritePhyUshort(sc, 0x02, 0x2010);
			MP_WritePhyUshort(sc, 0x03, 0xdc00);
			MP_WritePhyUshort(sc, 0x1f, 0x0000);
			MP_WritePhyUshort(sc, 0x0b, 0x0600);
			MP_WritePhyUshort(sc, 0x1f, 0x0005);
			MP_WritePhyUshort(sc, 0x05, 0xfff6);
			MP_WritePhyUshort(sc, 0x06, 0x00fc);
			MP_WritePhyUshort(sc, 0x1f, 0x0000);
		}

	}
	MP_WritePhyUshort(sc, 0x1F, 0x0000);

	anar = RL_PHY_100_FULL |
	       RL_PHY_100_HALF |
	       RL_PHY_10_FULL |
	       RL_PHY_10_HALF;

	gbcr = RL_PHY_1000_FULL |
	       RL_PHY_1000_HALF;

	bmcr = RL_PHY_NWAY_EN | RL_PHY_RESTART_NWAY;

	rl_set_speed_and_duplex(sc, gbcr, anar, bmcr);
}

void
MP_WritePhyUshort(sc,RegAddr,RegData)
	struct rl_softc		*sc;
	u_int8_t				RegAddr;
	u_int16_t				RegData;
{
	u_int32_t		TmpUlong=0x80000000;
	u_int32_t		Timeout=0;

	TmpUlong |= ( ((u_int32_t)RegAddr<<16) | (u_int32_t)RegData );

	CSR_WRITE_4(sc, PHYAR, TmpUlong );

	/* Wait for writing to Phy ok */
	do {
		Timeout++;
		DELAY( 20 );
		TmpUlong=CSR_READ_4(sc, PHYAR);
	} while ( (TmpUlong & PHYAR_Flag) && (Timeout<100000) );

}

u_int16_t
MP_ReadPhyUshort(sc,RegAddr)
	struct rl_softc		*sc;
	u_int8_t				RegAddr;
{
	u_int16_t		RegData;
	u_int32_t		TmpUlong=0x00000000;
	u_int32_t		Timeout=0;

	TmpUlong |= ( (u_int32_t)RegAddr << 16);
	CSR_WRITE_4(sc, PHYAR, TmpUlong );

	/* Wait for reading from Phy ok*/
	do {
		Timeout++;
		DELAY( 20 );
		TmpUlong=CSR_READ_4(sc, PHYAR);
	} while ( (!(TmpUlong & PHYAR_Flag)) && (Timeout<100000) );

	TmpUlong=CSR_READ_4(sc, PHYAR);
	RegData = (u_int16_t)(TmpUlong & 0x0000ffff);

	return RegData;
}

void MP_WriteEPhyUshort(struct rl_softc *sc, u_int8_t RegAddr, u_int16_t RegData)
{
	u_int32_t		TmpUlong=0x80000000;
	u_int32_t		Timeout=0;

	TmpUlong |= ( ((u_int32_t)RegAddr<<16) | (u_int32_t)RegData );

	CSR_WRITE_4(sc, RL_EPHYAR, TmpUlong );

	/* Wait for writing to Phy ok */
	do {
		Timeout++;
		DELAY( 100 );
		TmpUlong=CSR_READ_4(sc, RL_EPHYAR);
	} while ( (TmpUlong & PHYAR_Flag) && (Timeout<100) );

}

u_int16_t MP_ReadEPhyUshort(struct rl_softc *sc, u_int8_t RegAddr)
{
	u_int16_t		RegData;
	u_int32_t		TmpUlong;
	u_int32_t		Timeout=0;

	TmpUlong = ( (u_int32_t)RegAddr << 16);
	CSR_WRITE_4(sc, RL_EPHYAR, TmpUlong );

	// Wait for reading from EPhy ok
	do {
		Timeout++;
		DELAY( 100 );
		TmpUlong = CSR_READ_4(sc, RL_EPHYAR);
	} while ( (!(TmpUlong & PHYAR_Flag)) && (Timeout<100) );

	TmpUlong=CSR_READ_4(sc, RL_EPHYAR);
	RegData = (u_int16_t)(TmpUlong & 0x0000ffff);

	return RegData;
}

u_int8_t MP_ReadEfuse(struct rl_softc *sc, u_int16_t RegAddr)
{
	u_int8_t		RegData;
	u_int32_t		TmpUlong;
	u_int32_t		Timeout=0;

	RegAddr &= 0x3FF;
	TmpUlong = ( (u_int32_t)RegAddr << 8);
	CSR_WRITE_4(sc, 0xDC, TmpUlong );

	/* Wait for reading from Phy ok*/
	do {
		Timeout++;
		DELAY( 100 );
		TmpUlong = CSR_READ_4(sc, 0xDC);
	} while ( (!(TmpUlong & PHYAR_Flag)) && (Timeout<100) );

	TmpUlong=CSR_READ_4(sc, 0xDC);
	RegData = (u_int8_t)(TmpUlong & 0x000000ff);

	return RegData;
}

/*----------------------------------------------------------------------------*/
/*	8139 (CR9346) 9346 command register bits (offset 0x50, 1 byte)*/
/*----------------------------------------------------------------------------*/
#define CR9346_EEDO				0x01			/* 9346 data out*/
#define CR9346_EEDI				0x02			/* 9346 data in*/
#define CR9346_EESK				0x04			/* 9346 serial clock*/
#define CR9346_EECS				0x08			/* 9346 chip select*/
#define CR9346_EEM0				0x40			/* select 8139 operating mode*/
#define CR9346_EEM1				0x80			/* 00: normal*/
#define CR9346_CFGRW			0xC0			/* Config register write*/
#define CR9346_NORM			0x00

/*----------------------------------------------------------------------------*/
/*	EEPROM bit definitions(EEPROM control register bits)*/
/*----------------------------------------------------------------------------*/
#define EN_TRNF					0x10			/* Enable turnoff*/
#define EEDO						CR9346_EEDO	/* EEPROM data out*/
#define EEDI						CR9346_EEDI		/* EEPROM data in (set for writing data)*/
#define EECS						CR9346_EECS		/* EEPROM chip select (1=high, 0=low)*/
#define EESK						CR9346_EESK		/* EEPROM shift clock (1=high, 0=low)*/

/*----------------------------------------------------------------------------*/
/*	EEPROM opcodes*/
/*----------------------------------------------------------------------------*/
#define EEPROM_READ_OPCODE	06
#define EEPROM_WRITE_OPCODE	05
#define EEPROM_ERASE_OPCODE	07
#define EEPROM_EWEN_OPCODE	19				/* Erase/write enable*/
#define EEPROM_EWDS_OPCODE	16				/* Erase/write disable*/

#define	CLOCK_RATE				50				/* us*/

#define RaiseClock(_sc,_x)				\
	(_x) = (_x) | EESK;					\
	CSR_WRITE_1((_sc), RL_EECMD, (_x));	\
	DELAY(CLOCK_RATE);

#define LowerClock(_sc,_x)				\
	(_x) = (_x) & ~EESK;					\
	CSR_WRITE_1((_sc), RL_EECMD, (_x));	\
	DELAY(CLOCK_RATE);

/*
 * Shift out bit(s) to the EEPROM.
 */
static void rl_eeprom_ShiftOutBits(sc, data, count)
	struct rl_softc		*sc;
	int			data;
	int 			count;
{
	u_int16_t x,mask;

	mask = 0x01 << (count - 1);
	x = CSR_READ_1(sc, RL_EECMD);

	x &= ~(EEDO | EEDI);

	do
	{
		x &= ~EEDI;
		if(data & mask)
			x |= EEDI;

		CSR_WRITE_1(sc, RL_EECMD, x);
		DELAY(CLOCK_RATE);
		RaiseClock(sc,x);
		LowerClock(sc,x);
		mask = mask >> 1;
	} while(mask);

	x &= ~EEDI;
	CSR_WRITE_1(sc, RL_EECMD, x);
}

/*
 * Shift in bit(s) from the EEPROM.
 */
static u_int16_t rl_eeprom_ShiftInBits(sc)
	struct rl_softc		*sc;
{
	u_int16_t x,d,i;
	x = CSR_READ_1(sc, RL_EECMD);

	x &= ~( EEDO | EEDI);
	d = 0;

	for(i=0; i<16; i++)
	{
		d = d << 1;
		RaiseClock(sc, x);

		x = CSR_READ_1(sc, RL_EECMD);

		x &= ~(EEDI);
		if(x & EEDO)
			d |= 1;

		LowerClock(sc, x);
	}

	return d;
}

/*
 * Clean up EEprom read/write setting
 */
static void rl_eeprom_EEpromCleanup(sc)
	struct rl_softc		*sc;
{
	u_int16_t x;
	x = CSR_READ_1(sc, RL_EECMD);

	x &= ~(EECS | EEDI);
	CSR_WRITE_1(sc, RL_EECMD, x);

	RaiseClock(sc, x);
	LowerClock(sc, x);
}

/*
 * Read a word of data stored in the EEPROM at address 'addr.'
 */
static void rl_eeprom_getword(sc, addr, dest)
	struct rl_softc		*sc;
	int			addr;
	u_int16_t		*dest;
{
	u_int16_t x;

	/* select EEPROM, reset bits, set EECS*/
	x = CSR_READ_1(sc, RL_EECMD);

	x &= ~(EEDI | EEDO | EESK | CR9346_EEM0);
	x |= CR9346_EEM1 | EECS;
	CSR_WRITE_1(sc, RL_EECMD, x);

	/* write the read opcode and register number in that order*/
	/* The opcode is 3bits in length, reg is 6 bits long*/
	rl_eeprom_ShiftOutBits(sc, EEPROM_READ_OPCODE, 3);

	if (CSR_READ_4(sc, RL_RXCFG) & RL_RXCFG_RX_9356SEL)
		rl_eeprom_ShiftOutBits(sc, addr,8);	/*93c56=8*/
	else
		rl_eeprom_ShiftOutBits(sc, addr,6);	/*93c46=6*/

	/* Now read the data (16 bits) in from the selected EEPROM word*/
	*dest=rl_eeprom_ShiftInBits(sc);

	rl_eeprom_EEpromCleanup(sc);
	return;
}

/*
 * Read a sequence of words from the EEPROM.
 */
static void rl_read_eeprom(sc, dest, off, cnt, swap)
	struct rl_softc		*sc;
	caddr_t			dest;
	int			off;
	int			cnt;
	int			swap;
{
	int			i;
	u_int16_t		word = 0, *ptr;

	for (i = 0; i < cnt; i++)
	{
		rl_eeprom_getword(sc, off + i, &word);
		ptr = (u_int16_t *)(dest + (i * 2));
		if (swap)
			*ptr = ntohs(word);
		else
			*ptr = word;
	}

	return;
}
